import hydra
import yaml
import torch
from lensless.utils.plot import plot_image
from lensless.utils.dataset import DiffuserCamMirflickr
from torch.utils.data import Subset
import numpy as np
import os
from lensless.utils.io import save_image
import time
from lensless import ADMM
from lensless.recon.model_dict import load_model, download_model


repo_path = "/home/bezzam/LenslessPiCam"


@hydra.main(
    version_base=None, config_path="../../configs/recon", config_name="recon_diffusercam_mirflickr"
)
def apply_model(config):
    idx = config.idx
    save = config.save
    device = config.device
    n_trials = config.n_trials
    legacy_denoiser = config.legacy_denoiser

    # load dataset
    dataset = DiffuserCamMirflickr(
        dataset_dir=config.files.dataset,
        psf_path=os.path.join(repo_path, config.files.psf),
        downsample=config.files.downsample,
    )
    test_indices = dataset.allowed_idx[dataset.allowed_idx <= 1000]
    test_set = Subset(dataset, test_indices)
    psf = dataset.psf.to(device)
    print("Test set size:", len(test_set))
    print(f"Data shape :  {test_set[0][0].shape}")

    # load model
    model_name = config.model_name
    if model_name is None:
        print("Using traditional ADMM")
        recon = ADMM(psf, **config.admm)

    else:

        # load config
        model_path = download_model(camera="diffusercam", dataset="mirflickr", model=model_name)
        config_path = os.path.join(model_path, ".hydra", "config.yaml")
        with open(config_path, "r") as stream:
            model_config = yaml.safe_load(stream)

        # -- set seed
        seed = model_config["seed"]
        torch.manual_seed(seed)
        np.random.seed(seed)

        # load model
        recon = load_model(model_path, psf=psf, device=device, legacy_denoiser=legacy_denoiser)

    # apply reconstruction
    lensless, lensed = test_set[idx]

    start_time = time.time()
    for _ in range(n_trials):
        with torch.no_grad():

            recon.set_data(lensless.to(device))
            res = recon.apply(
                disp_iter=-1,
                save=False,
                gamma=None,
                plot=False,
            )
    end_time = time.time()
    avg_time_ms = (end_time - start_time) / n_trials * 1000
    print(f"Avg inference [ms] : {avg_time_ms:.3} ms")

    res_np = res[0].cpu().numpy().squeeze()

    plot_image(res_np)
    plot_image(lensed)

    if save:
        print(f"Saving images to {os.getcwd()}")
        lensed_np = lensed[0].cpu().numpy()
        save_image(lensed_np, f"{idx}_original_idx.png")
        save_image(res_np, f"{idx}_reconstruction_idx.png")
        save_image(lensless[0].cpu().numpy(), f"{idx}_lensless.png")


if __name__ == "__main__":
    apply_model()



import numpy as np
import os
from lensless import VirtualSensor, SensorOptions
from lensless.utils.io import load_image
from lensless.utils.image import rgb2gray


def test_sensor_size():
    for sensor_name in SensorOptions.values():
        sensor = VirtualSensor.from_name(sensor_name)
        if np.all(sensor.pixel_size * sensor.resolution > sensor.size):
            raise ValueError(
                f"Discrepancy in sensor {sensor_name} size. Pixel size * resolution = {sensor.pixel_size * sensor.resolution}, while size = {sensor.size}"
            )


def test_sensor_landscape():
    for sensor_name in SensorOptions.values():
        sensor = VirtualSensor.from_name(sensor_name)
        assert (
            sensor.resolution[0] <= sensor.resolution[1]
        ), f"Sensor {sensor_name} is not landscape."


def test_virtual_image():
    for sensor_name in SensorOptions.values():
        sensor = VirtualSensor.from_name(sensor_name)

        for bit_depth in sensor.bit_depth:
            img = sensor.capture()
            assert np.all(
                img.shape == sensor.image_shape
            ), f"Sensor {sensor_name} image shape is {img.shape}, while expected shape is {sensor.image_shape}."
            assert (
                img.max() <= 2**bit_depth - 1
            ), f"Sensor {sensor_name} image max value is {img.max()}, while expected max value is {2**bit_depth - 1}."


def test_virtual_image_from_rgb_file():
    """
    rgb > rgb
    rgb > gray
    """
    fp = os.path.join(os.path.dirname(__file__), "..", "data", "original", "tree.png")
    for sensor_name in SensorOptions.values():
        sensor = VirtualSensor.from_name(sensor_name)

        for bit_depth in sensor.bit_depth:
            img = sensor.capture(scene=fp, bit_depth=bit_depth)
            assert np.all(
                img.shape == sensor.image_shape
            ), f"Sensor {sensor_name} image shape is {img.shape}, while expected shape is {sensor.image_shape}."
            assert (
                img.max() <= 2**bit_depth - 1
            ), f"Sensor {sensor_name} image max value is {img.max()}, while expected max value is {2**bit_depth - 1}."


def test_virtual_image_from_gray_file():
    """
    gray > rgb
    gray > gray
    """
    fp = os.path.join(os.path.dirname(__file__), "..", "data", "original", "mnist_3.png")
    for sensor_name in SensorOptions.values():
        sensor = VirtualSensor.from_name(sensor_name)

        for bit_depth in sensor.bit_depth:
            img = sensor.capture(scene=fp, bit_depth=bit_depth)
            assert np.all(
                img.shape == sensor.image_shape
            ), f"Sensor {sensor_name} image shape is {img.shape}, while expected shape is {sensor.image_shape}."
            assert (
                img.max() <= 2**bit_depth - 1
            ), f"Sensor {sensor_name} image max value is {img.max()}, while expected max value is {2**bit_depth - 1}."


def test_virtual_image_from_rgb_data(save=False):
    fp = os.path.join(os.path.dirname(__file__), "..", "data", "original", "tree.png")

    # provided normalized float data
    img = load_image(fp)
    img = img.astype(np.float32)
    img /= img.max()
    for sensor_name in SensorOptions.values():
        sensor = VirtualSensor.from_name(sensor_name)

        for bit_depth in sensor.bit_depth:
            img_cap = sensor.capture(scene=img, bit_depth=bit_depth)
            assert np.all(
                img_cap.shape == sensor.image_shape
            ), f"Sensor {sensor_name} image shape is {img_cap.shape}, while expected shape is {sensor.image_shape}."
            assert (
                img_cap.max() <= 2**bit_depth - 1
            ), f"Sensor {sensor_name} image max value is {img_cap.max()}, while expected max value is {2**bit_depth - 1}."

        # save file
        if save:
            from lensless.utils.io import save_image

            save_image(img_cap, f"test_{sensor_name}.png")
            print(sensor_name, img_cap.shape)


def test_virtual_image_from_gray_data():
    fp = os.path.join(os.path.dirname(__file__), "..", "data", "original", "tree.png")

    # provided normalized float data
    img = load_image(fp)
    img = img.astype(np.float32)
    img /= img.max()
    img = rgb2gray(img, keepchanneldim=False)
    for sensor_name in SensorOptions.values():
        sensor = VirtualSensor.from_name(sensor_name)

        for bit_depth in sensor.bit_depth:
            img_cap = sensor.capture(scene=img, bit_depth=bit_depth)
            assert np.all(
                img_cap.shape == sensor.image_shape
            ), f"Sensor {sensor_name} image shape is {img_cap.shape}, while expected shape is {sensor.image_shape}."
            assert (
                img_cap.max() <= 2**bit_depth - 1
            ), f"Sensor {sensor_name} image max value is {img_cap.max()}, while expected max value is {2**bit_depth - 1}."


def test_downsample():
    fp = os.path.join(os.path.dirname(__file__), "..", "data", "original", "tree.png")
    downsample = 4

    for sensor_name in SensorOptions.values():
        sensor = VirtualSensor.from_name(sensor_name)
        new_res = (sensor.resolution / downsample).astype(int)
        sensor.downsample(downsample)

        img = sensor.capture(scene=fp)
        assert np.all(
            img.shape == sensor.image_shape
        ), f"Sensor {sensor_name} image shape is {img.shape}, while expected shape is {sensor.image_shape}."
        assert np.all(
            img.shape[:2] == new_res
        ), f"Sensor {sensor_name} image shape is {img.shape}, while expected shape is {new_res}."


if __name__ == "__main__":
    test_sensor_size()
    test_sensor_landscape()
    test_virtual_image()
    test_virtual_image_from_rgb_file()
    test_virtual_image_from_gray_file()
    test_virtual_image_from_rgb_data(save=True)
    test_virtual_image_from_gray_data()
    test_downsample()



"""

Load and plot measured PSF along with other useful analysis: pixel value
histogram and width.

Analyze PSF of lensless camera, namely looking at autocorrelations:
```python
python scripts/measure/analyze_image.py --fp data/psf/tape_rgb.png --gamma 2.2 --lensless
```

Example usage (data can be downloaded here: https://drive.switch.ch/index.php/s/NdgHlcDeHVDH5ww)
```bash
python scripts/measure/analyze_image.py --fp data/psf/lens_cardboard.png --plot_width 100 --lens
python scripts/measure/analyze_image.py --fp data/psf/lens_iris.png --plot_width 100 --lens
```

For Bayer data
```bash
python scripts/measure/analyze_image.py --fp data/psf/tape_bayer.png --bayer \
--gamma 2.2 --rg 2.1 --bg 1.3
```

To plot autocorrelations of lensless camera PSF
```bash
python scripts/measure/analyze_image.py --fp data/psf/tape_bayer.png --bayer \
--gamma 2.2 --rg 2.1 --bg 1.3 --lensless
```

Save RGB data from bayer
```
python scripts/measure/analyze_image.py --fp data/psf/tape_bayer.png --bayer \
--gamma 2.2 --rg 2.1 --bg 1.3 --save data/psf/tape_rgb.png
python scripts/measure/analyze_image.py --fp data/raw_data/thumbs_up_bayer.png --bayer \
--gamma 2.2 --rg 2.1 --bg 1.3 --save data/raw_data/thumbs_up_rgb.png
```

"""


import click
import cv2
import os
import numpy as np
import matplotlib.pyplot as plt
from lensless.utils.image import rgb2gray, gamma_correction, resize
from lensless.utils.plot import plot_image, pixel_histogram, plot_cross_section, plot_autocorr2d
from lensless.utils.io import load_image, load_psf, save_image


@click.command()
@click.option(
    "--fp",
    type=str,
    help="File path for measurement.",
)
@click.option(
    "--gamma",
    default=2.2,
    type=float,
    help="Gamma factor for plotting.",
)
@click.option(
    "--width",
    default=3,
    type=float,
    help="dB drop for estimating width",
)
@click.option(
    "--bayer",
    is_flag=True,
    help="Whether image is raw bayer data.",
)
@click.option(
    "--lens",
    is_flag=True,
    help="Whether measurement is PSF of lens, in that case plot cross-section of PSF.",
)
@click.option(
    "--lensless",
    is_flag=True,
    help="Whether measurement is PSF of a lensless camera, in that case plot cross-section of autocorrelation.",
)
@click.option(
    "--bg",
    type=float,
    help="Blue gain.",
)
@click.option(
    "--rg",
    type=float,
    help="Red gain.",
)
@click.option(
    "--plot_width",
    type=int,
    help="Width for cross-section.",
)
@click.option(
    "--save",
    type=str,
    help="File name to save color correct bayer as RGB.",
)
@click.option(
    "--save_auto",
    is_flag=True,
    help="Save autocorrelation instead of pop-up window.",
)
@click.option(
    "--nbits",
    default=None,
    type=int,
    help="Number of bits for output. Only used for Bayer data",
)
@click.option(
    "--down",
    default=1,
    type=int,
    help="Factor by which to downsample.",
)
@click.option(
    "--back",
    type=str,
    help="File path for background image, e.g. for screen.",
)
def analyze_image(
    fp, gamma, width, bayer, lens, lensless, bg, rg, plot_width, save, save_auto, nbits, down, back
):
    assert fp is not None, "Must pass file path."

    # initialize plotting axis
    fig_rgb, ax_rgb = plt.subplots(ncols=2, nrows=1, num="RGB", figsize=(15, 5))
    if lens:
        fig_gray, ax_gray = plt.subplots(ncols=3, nrows=1, num="Grayscale", figsize=(15, 5))
    else:
        fig_gray, ax_gray = plt.subplots(ncols=2, nrows=1, num="Grayscale", figsize=(15, 5))

    # load PSF/image
    if lensless:
        img = load_psf(
            fp,
            verbose=True,
            bayer=bayer,
            blue_gain=bg,
            red_gain=rg,
            nbits_out=nbits,
            return_float=False,
            downsample=down,
        )[0]
    else:
        img = load_image(
            fp,
            verbose=True,
            bayer=bayer,
            blue_gain=bg,
            red_gain=rg,
            nbits_out=nbits,
            back=back,
            downsample=down,
        )
    if nbits is None:
        nbits = int(np.ceil(np.log2(img.max())))
    # plot RGB and grayscale
    ax = plot_image(img, gamma=gamma, normalize=True, ax=ax_rgb[0])
    ax.set_title("RGB")
    ax = pixel_histogram(img, ax=ax_rgb[1], nbits=nbits)
    ax.set_title("Histogram")
    fig_rgb.savefig(os.path.join(os.path.dirname(fp), "rgb_analysis.png"))

    img_grey = rgb2gray(img[None, ...])
    ax = plot_image(img_grey, gamma=gamma, normalize=True, ax=ax_gray[0])
    ax.set_title("Grayscale")
    ax = pixel_histogram(img_grey, ax=ax_gray[1], nbits=nbits)
    ax.set_title("Histogram")
    fig_gray.savefig(os.path.join(os.path.dirname(fp), "grey_analysis.png"))

    img_grey = img_grey.squeeze()
    img = img.squeeze()

    if lens:
        # determine PSF width
        plot_cross_section(
            img_grey, color="gray", plot_db_drop=width, ax=ax_gray[2], plot_width=plot_width
        )
        fig_auto, ax_cross = plt.subplots(ncols=3, nrows=1, num="RGB widths", figsize=(15, 5))
        for i, c in enumerate(["r", "g", "b"]):
            print(f"-- {c} channel")
            ax, _ = plot_cross_section(
                img[:, :, i],
                color=c,
                ax=ax_cross[i],
                plot_db_drop=width,
                max_val=2**nbits - 1,
                plot_width=plot_width,
            )
            if i > 0:
                ax.set_ylabel("")

    elif lensless:
        # plot autocorrelations and width
        # -- grey
        fig_auto, ax_auto = plt.subplots(ncols=4, nrows=2, num="Autocorrelations", figsize=(15, 5))
        _, autocorr_grey = plot_autocorr2d(img_grey, ax=ax_auto[0][0])
        print("-- grayscale")
        plot_cross_section(
            autocorr_grey, color="gray", plot_db_drop=width, ax=ax_auto[1][0], plot_width=plot_width
        )
        # -- rgb
        for i, c in enumerate(["r", "g", "b"]):
            _, autocorr_c = plot_autocorr2d(img[:, :, i], ax=ax_auto[0][i + 1])
            print(f"-- {c} channel")
            ax, _ = plot_cross_section(
                autocorr_c,
                color=c,
                ax=ax_auto[1][i + 1],
                plot_db_drop=width,
                plot_width=plot_width,
            )
            ax.set_ylabel("")

    if bayer and save is not None:
        cv2.imwrite(save, cv2.cvtColor(img, cv2.COLOR_RGB2BGR))
        print(f"\nColor-corrected RGB image saved to: {save}")

        # save 8bit version for visualization
        if gamma is not None:
            img = img / img.max()
            img = gamma_correction(img, gamma=gamma)
        save_8bit = save.replace(".png", "_8bit.png")
        save_image(img, save_8bit, normalize=True)
        print(f"\n8bit version saved to: {save_8bit}")

    if save_auto:
        auto_fp = os.path.join(os.path.dirname(fp), "autocorrelation.png")
        fig_auto.savefig(auto_fp)
        print(f"\nAutocorrelation saved to: {auto_fp}")
    else:
        plt.show()


if __name__ == "__main__":
    analyze_image()



"""

Telegram bot to interface with lensless camera setup.

"""

import hydra
import logging
import numpy as np
import os
from PIL import Image, ImageFont
import shutil
import pytz
from datetime import datetime
from lensless.hardware.utils import check_username_hostname
from lensless.hardware.slm import set_programmable_mask, adafruit_sub2full

# for displaying emojis
from emoji import EMOJI_DATA
from lensless.utils.io import load_psf
from pilmoji import Pilmoji

from telegram import __version__ as TG_VER

try:
    from telegram import __version_info__
except ImportError:
    __version_info__ = (0, 0, 0, 0, 0)  # type: ignore[assignment]

if __version_info__ < (20, 0, 0, "alpha", 1):
    raise RuntimeError(
        f"This example is not compatible with your current PTB version {TG_VER}. To view the "
        f"{TG_VER} version of this example, "
        f"visit https://docs.python-telegram-bot.org/en/v{TG_VER}/examples.html"
    )
from telegram.ext import (
    Application,
    CommandHandler,
    ContextTypes,
    MessageHandler,
    filters,
    CallbackQueryHandler,
)
from telegram import Update, InlineKeyboardButton, InlineKeyboardMarkup


TOKEN = None
RPI_USERNAME = None
RPI_HOSTNAME = None
RPI_LENSED_USERNAME = None
RPI_LENSED_HOSTNAME = None
CONFIG_FN = None
DEFAULT_ALGO = None
ALGO_TEXT = None
MASK_PARAM = None
TIME_OFFSET = None


OVERLAY_ALPHA = None
OVERLAY_1 = None
OVERLAY_2 = None
OVERLAY_3 = None

SETUP_FP = "docs/source/demo_setup.png"
INPUT_FP = "user_photo.jpg"
RAW_DATA_FP = "raw_data.png"
OUTPUT_FOLDER = "demo_lensless"
BUSY = False
# supported_algos = ["fista", "admm", "unrolled"]
supported_algos = ["fista", "admm"]
supported_input = ["mnist", "thumb", "face"]
FILES_CAPTURE_CONFIG = None
TIMEOUT = 1 * 60  # 10 minutes

BRIGHTNESS = 80
EXPOSURE = 0.02
LOW_LIGHT_THRESHOLD = 100
SATURATION_THRESHOLD = 0.05

PSF_FP = None
PSF_FP_GAMMA = os.path.join(OUTPUT_FOLDER, "psf_gamma.png")
BACKGROUND_FP = None

MAX_QUERIES_PER_DAY = 20
queries_count = dict()
WHITELIST_USERS = None


# Enable logging
logging.basicConfig(
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


def check_algo(algo):

    if algo not in supported_algos:
        return False
    else:
        return True


def get_user_folder(update):
    # name = update.message.from_user.full_name.replace(" ", "-")
    # user_subfolder = f"{update.message.from_user.id}_{name}"
    user_subfolder = f"{update.message.from_user.id}"
    return os.path.join(OUTPUT_FOLDER, user_subfolder)


def get_user_folder_from_query(query):
    # name = query.message.from_user.full_name.replace(" ", "-")
    # user_subfolder = f"{query.message.from_user.id}_{name}"
    user_subfolder = f"{query.message.from_user.id}"
    return os.path.join(OUTPUT_FOLDER, user_subfolder)


async def remove_busy_flag(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
    global BUSY
    BUSY = False


async def check_incoming_message(update: Update, context: ContextTypes.DEFAULT_TYPE):

    global BUSY, queries_count

    # create folder for user
    user_folder = get_user_folder(update)
    if not os.path.exists(user_folder):
        os.makedirs(user_folder, exist_ok=True)

        if MASK_PARAM is not None:

            import torch
            from lensless.hardware.slm import set_programmable_mask, adafruit_sub2full
            from lensless.hardware.trainable_mask import AdafruitLCD
            from lensless.utils.io import save_image

            user_id = int(os.path.basename(user_folder))
            np.random.seed(user_id % (2**32 - 1))  # TODO set user ID as seed
            mask_vals = np.random.uniform(0, 1, MASK_PARAM.mask_shape)

            # simulate PSF
            full_pattern = adafruit_sub2full(
                mask_vals,
                center=MASK_PARAM.mask_center,
            )
            set_programmable_mask(
                full_pattern,
                device=MASK_PARAM.device,
                rpi_username=RPI_USERNAME,
                rpi_hostname=RPI_HOSTNAME,
            )
            mask_vals_torch = torch.from_numpy(mask_vals.astype(np.float32))
            mask = AdafruitLCD(
                initial_vals=mask_vals_torch,
                sensor=MASK_PARAM.sensor,
                slm=MASK_PARAM.device,
                downsample=MASK_PARAM.downsample,
                flipud=MASK_PARAM.flipud,
            )
            psf = mask.get_psf().detach().numpy()

            # save PSF as PNG
            psf_fp = os.path.join(user_folder, "psf.png")
            save_image(psf[0], psf_fp)

            # save as NPY
            psf_npy_fp = os.path.join(user_folder, "psf.npy")
            np.save(psf_npy_fp, psf)

    if BUSY:
        return "System is busy. Please wait for the current job to finish and try again."

    # if message from a while ago, ignore
    utc = pytz.UTC
    now = utc.localize(datetime.now())
    message_time = update.message.date

    diff = (now - message_time).total_seconds()
    diff -= TIME_OFFSET

    if diff > TIMEOUT:
        return f"Timeout ({TIMEOUT} seconds) exceeded. Someone else may be using the system. Please send a new message."

    if len(update.message.photo) > 1:
        original_file_path = os.path.join(user_folder, INPUT_FP)
        photo_file = await update.message.photo[-1].get_file()
        await photo_file.download_to_drive(original_file_path)
        img = np.array(Image.open(original_file_path))

        # -- check if portrait
        if img.shape[0] < img.shape[1]:
            return "Please send a portrait photo."
            # await update.message.reply_text("Please send a portrait photo.", reply_to_message_id=update.message.message_id)
            # return
        else:
            await update.message.reply_text(
                "Got photo of resolution: " + str(img.shape),
                reply_to_message_id=update.message.message_id,
            )

    # check that not command
    elif update.message.text[0] != "/" and len(update.message.text) > 0:

        text = update.message.text

        if len(update.message.text) > 1 or text not in EMOJI_DATA:
            return "Supported text for display is only a single emoji."

    # increment queries count
    user_id = update.message.from_user.id
    if user_id not in queries_count:
        queries_count[user_id] = 1
    else:
        queries_count[user_id] += 1
        if user_id not in WHITELIST_USERS:
            if queries_count[user_id] > MAX_QUERIES_PER_DAY:
                return f"Maximum number of queries per day ({MAX_QUERIES_PER_DAY}) exceeded. Please try again tomorrow."
            return

    # print user
    print("User: ", update.message.from_user, update.message.from_user.id)
    print("Queries count: ", queries_count[user_id])

    # reset at midnight
    if now.hour == 0 and now.minute == 0:
        queries_count = dict()

    BUSY = True

    return


# Define a few command handlers. These usually take the two arguments update and
# context.
async def start(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
    """Send a message when the command /start is issued."""
    user = update.effective_user
    # await update.message.reply_html(
    #     ,
    # )

    await update.message.reply_photo(SETUP_FP, caption=f"Hi {user.first_name}! " + HELP_TEXT)
    await update.message.reply_text(ALGO_TEXT)


async def help_command(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
    """Send a message when the command /help is issued."""
    await update.message.reply_photo(SETUP_FP, caption=HELP_TEXT)
    await update.message.reply_text(ALGO_TEXT)


async def fista(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:

    await reconstruct(update, context, algo="fista")


async def admm(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:

    await reconstruct(update, context, algo="admm")


async def unrolled(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:

    await reconstruct(update, context, algo="unrolled")


async def unet(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:

    await reconstruct(update, context, algo="unet")


async def photo(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:

    """
    1. Get photo from user
    2. Send to display
    3. Capture measurement
    4. Reconstruct
    """

    global BUSY, EXPOSURE

    res = await check_incoming_message(update, context)
    if res is not None:
        await update.message.reply_text(res, reply_to_message_id=update.message.message_id)
        return

    algo = update.message.caption
    if algo is not None:
        algo = algo.lower()
    else:
        algo = DEFAULT_ALGO

    if check_algo(algo):

        # # call python script for full process
        # os.system(f"python scripts/demo.py plot=False fp={INPUT_FP} output={OUTPUT_FOLDER}")

        # -- send to display
        user_folder = get_user_folder(update)
        original_file_path = os.path.join(user_folder, INPUT_FP)
        os.system(
            f"python scripts/measure/remote_display.py -cn {CONFIG_FN} fp={original_file_path} rpi.username={RPI_USERNAME} rpi.hostname={RPI_HOSTNAME}"
        )
        await update.message.reply_text(
            "Image sent to display.", reply_to_message_id=update.message.message_id
        )

        await take_picture_and_reconstruct(update, context, algo)

    else:

        await update.message.reply_text(
            f"Unsupported algorithm: {algo}. Please specify from: {supported_algos}",
            reply_to_message_id=update.message.message_id,
        )

    BUSY = False


async def take_picture(update: Update, context: ContextTypes.DEFAULT_TYPE, query=None) -> None:

    # get user subfolder
    if query is not None:
        user_subfolder = get_user_folder_from_query(query)
        await query.message.reply_text(
            f"Taking picture with exposure of {EXPOSURE} seconds...",
            reply_to_message_id=query.message.message_id,
        )
    else:
        user_subfolder = get_user_folder(update)
        await update.message.reply_text(
            f"Taking picture with exposure of {EXPOSURE} seconds...",
            reply_to_message_id=update.message.message_id,
        )

    os.system(
        f"python scripts/measure/remote_capture.py -cn {CONFIG_FN} plot=False rpi.username={RPI_USERNAME} rpi.hostname={RPI_HOSTNAME} output={user_subfolder} capture.exp={EXPOSURE}"
    )


def overlay(user_subfolder):

    if OVERLAY_1 is not None or OVERLAY_2 is not None or OVERLAY_3 is not None:

        alpha = OVERLAY_ALPHA

        reconstructed_path = os.path.join(user_subfolder, "reconstructed.png")

        img1 = Image.open(reconstructed_path)
        img1 = img1.convert("RGBA")

        for overlay_config in [OVERLAY_1, OVERLAY_2, OVERLAY_3]:
            if overlay_config is not None:
                overlay_img = Image.open(overlay_config.fp)
                overlay_img = overlay_img.convert("RGBA")
                overlay_img.putalpha(alpha)
                new_width = int(img1.width * overlay_config.scaling)
                overlay_img = overlay_img.resize(
                    (new_width, int(new_width * overlay_img.height / overlay_img.width))
                )
                img1.paste(
                    overlay_img,
                    (overlay_config.position[0], overlay_config.position[1]),
                    overlay_img,
                )

        OUTPUT_FP = os.path.join(user_subfolder, "reconstructed_overlay.png")
        img1.convert("RGB").save(OUTPUT_FP)

    else:

        OUTPUT_FP = os.path.join(user_subfolder, "reconstructed.png")

    return OUTPUT_FP


async def random_mask(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
    """
    Set random mask and reconstruct (with ADMM).
    """

    algo = "admm"

    # get user subfolder
    user_subfolder = get_user_folder(update)

    # get seed for random mask
    seed = int(os.path.basename(user_subfolder))
    # add random number to seed
    seed += np.random.randint(0, 1000)
    os.system(
        f"python scripts/recon/demo.py -cn {CONFIG_FN} plot=False recon.algo={algo} output={user_subfolder} camera.psf.seed={seed}"
    )

    # -- send back, with watermark if provided
    OUTPUT_FP = overlay(user_subfolder)
    await update.message.reply_photo(
        OUTPUT_FP,
        caption=f"Reconstruction ({algo})",
        reply_to_message_id=update.message.message_id,
    )

    # simulate BAD PSF
    import torch
    from lensless.hardware.slm import set_programmable_mask, adafruit_sub2full
    from lensless.hardware.trainable_mask import AdafruitLCD
    from lensless.utils.io import save_image

    np.random.seed(seed % (2**32 - 1))
    mask_vals = np.random.uniform(0, 1, MASK_PARAM.mask_shape)

    # simulate PSF
    full_pattern = adafruit_sub2full(
        mask_vals,
        center=MASK_PARAM.mask_center,
    )
    set_programmable_mask(
        full_pattern, device=MASK_PARAM.device, rpi_username=RPI_USERNAME, rpi_hostname=RPI_HOSTNAME
    )
    mask_vals_torch = torch.from_numpy(mask_vals.astype(np.float32))
    mask = AdafruitLCD(
        initial_vals=mask_vals_torch,
        sensor=MASK_PARAM.sensor,
        slm=MASK_PARAM.device,
        downsample=MASK_PARAM.downsample,
        flipud=MASK_PARAM.flipud,
    )
    psf = mask.get_psf().detach().numpy()

    # save PSF as PNG
    psf_fp = os.path.join(user_subfolder, "psf_bad.png")
    save_image(psf[0], psf_fp)
    await update.message.reply_photo(
        psf_fp,
        caption="Incorrect PSF used for reconstruction",
        reply_to_message_id=update.message.message_id,
    )

    # send back false and ground truth PSF
    psf_fp = os.path.join(user_subfolder, "psf.png")
    await update.message.reply_photo(
        psf_fp,
        caption="Correct PSF (your key)",
        reply_to_message_id=update.message.message_id,
    )


async def reconstruct(update: Update, context: ContextTypes.DEFAULT_TYPE, algo, query=None) -> None:

    supported = check_algo(algo)
    if not supported:
        await update.message.reply_text(
            f"Unsupported algorithm: {algo}. Please specify from: {supported_algos}",
            reply_to_message_id=update.message.message_id,
        )
        return

    # get user subfolder
    if query is not None:
        user_subfolder = get_user_folder_from_query(query)
        update = query  # to get the reply_to_message_id
    else:
        user_subfolder = get_user_folder(update)

    # check file exists
    raw_data = os.path.join(user_subfolder, RAW_DATA_FP)
    print(raw_data)
    if not os.path.exists(raw_data):
        await update.message.reply_text(
            "No data to reconstruct. Please take a picture first.",
            reply_to_message_id=update.message.message_id,
        )
        return

    await update.message.reply_text(
        f"Reconstructing with {algo}...", reply_to_message_id=update.message.message_id
    )
    if PSF_FP is not None:
        os.system(
            f"python scripts/recon/demo.py -cn {CONFIG_FN} plot=False recon.algo={algo} output={user_subfolder} camera.psf={PSF_FP} recon.downsample=1 camera.background={BACKGROUND_FP}"
        )
    elif MASK_PARAM is not None:
        # get seed for random mask
        seed = int(os.path.basename(user_subfolder))
        os.system(
            f"python scripts/recon/demo.py -cn {CONFIG_FN} plot=False recon.algo={algo} output={user_subfolder} camera.psf.seed={seed}"
        )
    else:
        os.system(
            f"python scripts/recon/demo.py -cn {CONFIG_FN} plot=False recon.algo={algo} output={user_subfolder}"
        )

    # -- send back, with watermark if provided
    OUTPUT_FP = overlay(user_subfolder)
    await update.message.reply_photo(
        OUTPUT_FP,
        caption=f"Reconstruction ({algo})",
        reply_to_message_id=update.message.message_id,
    )


async def take_picture_and_reconstruct(
    update: Update, context: ContextTypes.DEFAULT_TYPE, algo, query=None
) -> None:

    if query is not None:
        user_subfolder = get_user_folder_from_query(query)
        responder = query.message
    else:
        user_subfolder = get_user_folder(update)
        responder = update.message

    # (if DigiCam) set mask pattern
    if MASK_PARAM is not None:
        print("Setting mask pattern...")
        # get seed for random mask
        seed = int(os.path.basename(user_subfolder))
        np.random.seed(seed % (2**32 - 1))
        mask_vals = np.random.uniform(0, 1, MASK_PARAM.mask_shape)
        full_pattern = adafruit_sub2full(
            mask_vals,
            center=MASK_PARAM.mask_center,
        )
        # setting mask
        set_programmable_mask(
            full_pattern,
            device=MASK_PARAM.device,
            rpi_username=RPI_USERNAME,
            rpi_hostname=RPI_HOSTNAME,
        )

    await take_picture(update, context, query=query)

    # check for saturation
    OUTPUT_FP = os.path.join(user_subfolder, "raw_data_8bit.png")
    # -- load picture to check for saturation
    img = np.array(Image.open(OUTPUT_FP))
    ratio = np.sum(img == 255) / np.prod(img.shape)
    if ratio > SATURATION_THRESHOLD:

        if EXPOSURE > 0.02:
            warning_message = (
                "ERROR: saturation/clipping detected in raw measurement!"
                "\nTry reducing the /brightness of the screen, or the /exposure of the camera."
            )
        else:
            warning_message = (
                "ERROR: saturation/clipping detected in raw measurement!"
                "\nTry reducing the /brightness of the screen."
            )
        await responder.reply_photo(
            OUTPUT_FP, caption=warning_message, reply_to_message_id=responder.message_id
        )
        return

    # -- reconstruct
    await reconstruct(update, context, algo, query=query)

    # # send picture of raw measurement
    # OUTPUT_FP = os.path.join(user_subfolder, "raw_data_8bit.png")
    # # -- load picture to check for saturation
    # img = np.array(Image.open(OUTPUT_FP))
    # ratio = np.sum(img == 255) / np.prod(img.shape)
    # if ratio > 0.05:

    #     if EXPOSURE > 0.02:
    #         warning_message = ("WARNING: saturation detected in raw measurement!"
    #             "\nTry reducing the /brightness of the screen, or the /exposure of the camera.")
    #     else:
    #         warning_message = ("WARNING: saturation detected in raw measurement!"
    #             "\nTry reducing the /brightness of the screen.")

    #     await responder.reply_photo(
    #         OUTPUT_FP,
    #         caption=warning_message,
    #         reply_to_message_id=responder.message_id
    #     )

    # # -- check if low light -> send warning to increase exposure
    # if np.max(img) < LOW_LIGHT_THRESHOLD:
    #     caption = "Raw measurement.\nWARNING: low light measurement. Try increasing the /exposure of the camera."
    # else:
    caption = "Raw measurement"

    # else:
    await responder.reply_photo(
        OUTPUT_FP, caption=caption, reply_to_message_id=responder.message_id
    )

    # -- send picture of setup (lensed)
    if RPI_LENSED_HOSTNAME is not None and RPI_LENSED_USERNAME is not None:
        os.system(
            f"python scripts/measure/remote_capture.py -cn {CONFIG_FN} rpi.username={RPI_LENSED_USERNAME} rpi.hostname={RPI_LENSED_HOSTNAME} plot=False capture.bayer=False capture.down=8 output={user_subfolder} capture.raw_data_fn=lensed capture.awb_gains=null"
        )
        OUTPUT_FP = os.path.join(user_subfolder, "lensed.png")
        await responder.reply_photo(
            OUTPUT_FP, caption="Picture of setup", reply_to_message_id=responder.message_id
        )


async def file_input_command(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:

    global BUSY, EXPOSURE

    res = await check_incoming_message(update, context)
    if res is not None:
        await update.message.reply_text(res, reply_to_message_id=update.message.message_id)
        return

    algo = DEFAULT_ALGO

    # extract config by based on file name
    file_name = update.message.text[1:]
    if file_name not in FILES_CAPTURE_CONFIG:
        await update.message.reply_text(
            f"Unsupported file: {file_name}. Please specify from: {FILES_CAPTURE_CONFIG.keys()}",
            reply_to_message_id=update.message.message_id,
        )
        return

    brightness = FILES_CAPTURE_CONFIG[file_name]["brightness"]
    EXPOSURE = FILES_CAPTURE_CONFIG[file_name]["exposure"]
    fp = FILES_CAPTURE_CONFIG[file_name]["fp"]

    # copy image to INPUT_FP
    user_folder = get_user_folder(update)
    original_file_path = os.path.join(user_folder, INPUT_FP)
    os.system(f"cp {fp} {original_file_path}")

    # -- send to display
    os.system(
        f"python scripts/measure/remote_display.py -cn {CONFIG_FN} fp={original_file_path} display.brightness={brightness} rpi.username={RPI_USERNAME} rpi.hostname={RPI_HOSTNAME}"
    )
    await update.message.reply_text(
        f"Image sent to display with brightness {brightness}.",
        reply_to_message_id=update.message.message_id,
    )

    await take_picture_and_reconstruct(update, context, algo)
    BUSY = False


async def psf_command(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:

    """
    Measure PSF through screen
    """

    # global BUSY

    res = await check_incoming_message(update, context)
    if res is not None:
        await update.message.reply_text(res, reply_to_message_id=update.message.message_id)
        return

    # vshift = -15
    # psf_size = 10

    # # -- send to display
    # os.system(
    #     f"python scripts/measure/remote_display.py -cn {CONFIG_FN} display.psf={psf_size} display.vshift={vshift} rpi.username={RPI_USERNAME} rpi.hostname={RPI_HOSTNAME}"
    # )
    # await update.message.reply_text(
    #     f"PSF of {psf_size}x{psf_size} pixels set on display.",
    #     reply_to_message_id=update.message.message_id,
    # )

    # # -- measurement
    # os.system(
    #     f"python scripts/measure/remote_capture.py -cn demo_measure_psf rpi.username={RPI_USERNAME} rpi.hostname={RPI_HOSTNAME}"
    # )
    # OUTPUT_FP = os.path.join(OUTPUT_FOLDER, "raw_data.png")
    # await update.message.reply_photo(
    #     OUTPUT_FP,
    #     caption="PSF (zoom in to see pattern)",
    #     reply_to_message_id=update.message.message_id,
    # )

    # send back ground truth PSF
    if PSF_FP is not None:
        await update.message.reply_photo(
            PSF_FP_GAMMA,
            caption="PSF used for reconstructions",
            reply_to_message_id=update.message.message_id,
        )
    elif MASK_PARAM is not None:
        # return pre-computed PSF
        user_folder = get_user_folder(update)
        psf_fp = os.path.join(user_folder, "psf.png")
        await update.message.reply_photo(
            psf_fp,
            caption="PSF used for reconstructions",
            reply_to_message_id=update.message.message_id,
        )

    # BUSY = False


async def button(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
    """Parses the CallbackQuery and updates the message text."""

    global BRIGHTNESS, EXPOSURE, BUSY
    BUSY = True

    query = update.callback_query

    # CallbackQueries need to be answered, even if no notification to the user is needed
    # Some clients may have trouble otherwise. See https://core.telegram.org/bots/api#callbackquery
    await query.answer()

    if query.data == "Cancel":
        BUSY = False
        await query.edit_message_text(text="Cancelled.")
        return

    if "brightness" in query.message.text:

        BRIGHTNESS = int(query.data)

        await query.edit_message_text(text=f"Screen brightness set to: {query.data}")

        # -- resend to display
        user_folder = get_user_folder_from_query(query)
        original_file_path = os.path.join(user_folder, INPUT_FP)
        os.system(
            f"python scripts/measure/remote_display.py -cn {CONFIG_FN} fp={original_file_path} display.brightness={BRIGHTNESS} rpi.username={RPI_USERNAME} rpi.hostname={RPI_HOSTNAME}"
        )
        await query.edit_message_text(text=f"Image sent to display with brightness {BRIGHTNESS}.")
        # await update.message.reply_text("Image sent to display.", reply_to_message_id=update.message.message_id)

    elif "exposure" in query.message.text:

        EXPOSURE = float(query.data)
        await query.edit_message_text(text=f"Exposure set to {EXPOSURE} seconds.")

    # TODO not working with mask
    algo = DEFAULT_ALGO
    # send query instead of update as it has the message data
    await take_picture_and_reconstruct(update, context, algo, query=query)
    BUSY = False


async def brightness_command(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:

    """
    Set brightness, re-capture, and reconstruct.
    """

    # check INPUT_FP exists
    user_folder = get_user_folder(update)
    original_file_path = os.path.join(user_folder, INPUT_FP)
    if not os.path.exists(original_file_path):
        await update.message.reply_text(
            "Please set an image first.", reply_to_message_id=update.message.message_id
        )
        return

    res = await check_incoming_message(update, context)
    if res is not None:
        await update.message.reply_text(res, reply_to_message_id=update.message.message_id)
        return

    vals = [20, 40, 60, 80, 100]
    vals.remove(BRIGHTNESS)
    keyboard = [
        [
            InlineKeyboardButton(f"{vals[0]}", callback_data=f"{vals[0]}"),
            InlineKeyboardButton(f"{vals[1]}", callback_data=f"{vals[1]}"),
        ],
        [
            InlineKeyboardButton(f"{vals[2]}", callback_data=f"{vals[2]}"),
            InlineKeyboardButton(f"{vals[3]}", callback_data=f"{vals[3]}"),
        ],
        [
            InlineKeyboardButton("Cancel", callback_data="Cancel"),
        ],
    ]

    reply_markup = InlineKeyboardMarkup(keyboard)

    await update.message.reply_text(
        f"Please specify a value for the screen brightness. Current value is {BRIGHTNESS}",
        # reply_to_message_id=update.message.message_id,
        # reply_markup=ReplyKeyboardMarkup(
        #     reply_keyboard, resize_keyboard=True, one_time_keyboard=True, is_persistent=False, input_field_placeholder=f"Screen brightness value (current={BRIGHTNESS})."
        # ),
        reply_markup=reply_markup,
    )


async def exposure_command(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:

    """
    Set exposure, re-capture, and reconstruct.
    """

    # # check INPUT_FP exists
    # user_folder = get_user_folder(update)
    # original_file_path = os.path.join(user_folder, INPUT_FP)
    # if not os.path.exists(original_file_path):
    #     await update.message.reply_text(
    #         "Please set an image first.", reply_to_message_id=update.message.message_id
    #     )
    #     return

    res = await check_incoming_message(update, context)
    if res is not None:
        await update.message.reply_text(res, reply_to_message_id=update.message.message_id)
        return

    # -- phase mask
    vals = {0.02: "very low", 0.04: "low", 0.06: "medium", 0.08: "high", 0.1: "very high"}
    # # -- tape based
    # vals = {0.02: "very low", 0.035: "low", 0.05: "medium", 0.065: "high", 0.08: "very high"}
    # # -- digicam
    # vals = {0.25: "very low", 0.5: "low", 0.75: "medium", 1: "high", 1.25: "very high"}

    if EXPOSURE in vals:
        del vals[EXPOSURE]
    keys = list(vals.keys())
    keyboard = [
        [
            InlineKeyboardButton(f"{vals[keys[0]]} ({keys[0]})", callback_data=f"{keys[0]}"),
            InlineKeyboardButton(f"{vals[keys[1]]} ({keys[1]})", callback_data=f"{keys[1]}"),
        ],
        [
            InlineKeyboardButton(f"{vals[keys[2]]} ({keys[2]})", callback_data=f"{keys[2]}"),
            InlineKeyboardButton(f"{vals[keys[3]]} ({keys[3]})", callback_data=f"{keys[3]}"),
        ],
        [
            InlineKeyboardButton("Cancel", callback_data="Cancel"),
        ],
    ]

    reply_markup = InlineKeyboardMarkup(keyboard)

    await update.message.reply_text(
        f"Please specify a value for the camera exposure. Current value is ({EXPOSURE} seconds).",
        reply_markup=reply_markup,
    )


async def emoji(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:

    global BUSY, EXPOSURE

    EXPOSURE = 0.7

    res = await check_incoming_message(update, context)
    if res is not None:
        await update.message.reply_text(res, reply_to_message_id=update.message.message_id)
        return

    # create image from emoji
    text = update.message.text
    size = 15
    with Image.new("RGB", (size, size), (0, 0, 0)) as image:
        font = ImageFont.truetype(
            "/usr/share/fonts/truetype/freefont/FreeMono.ttf", size, encoding="unic"
        )

        with Pilmoji(image) as pilmoji:
            pilmoji.text((0, 0), text.strip(), (0, 0, 0), font, align="center")

        # save image
        user_folder = get_user_folder(update)
        original_file_path = os.path.join(user_folder, INPUT_FP)
        image.save(original_file_path)

    # display
    vshift = -20
    brightness = 80
    os.system(
        f"python scripts/measure/remote_display.py -cn {CONFIG_FN} fp={original_file_path} rpi.username={RPI_USERNAME} rpi.hostname={RPI_HOSTNAME} display.vshift={vshift} display.brightness={brightness}"
    )
    await update.message.reply_text(
        f"Image sent to display with brightness {brightness}.",
        reply_to_message_id=update.message.message_id,
    )

    await take_picture_and_reconstruct(update, context, DEFAULT_ALGO)
    BUSY = False


async def not_running_command(update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
    await update.message.reply_text(
        "The bot is currently not running. If you want to try it out, please contact the admin.",
        reply_to_message_id=update.message.message_id,
    )


@hydra.main(version_base=None, config_path="../../configs", config_name="telegram_demo")
def main(config) -> None:
    """Start the bot."""

    global TOKEN, WHITELIST_USERS, RPI_USERNAME, RPI_HOSTNAME, RPI_LENSED_USERNAME, RPI_LENSED_HOSTNAME, CONFIG_FN, TIME_OFFSET
    global DEFAULT_ALGO, ALGO_TEXT, HELP_TEXT, supported_algos, supported_input
    global OVERLAY_ALPHA, OVERLAY_1, OVERLAY_2, OVERLAY_3, FILES_CAPTURE_CONFIG
    global PSF_FP, BACKGROUND_FP, MASK_PARAM, SETUP_FP
    global VSHIFT, IMAGE_RES

    TOKEN = config.token
    TIME_OFFSET = config.time_offset

    WHITELIST_USERS = config.whitelist
    if WHITELIST_USERS is None:
        WHITELIST_USERS = []

    if config.setup_fp is not None:
        SETUP_FP = config.setup_fp
        assert os.path.exists(SETUP_FP)

    RPI_USERNAME = config.rpi_username
    RPI_HOSTNAME = config.rpi_hostname
    RPI_LENSED_USERNAME = config.rpi_lensed_username
    RPI_LENSED_HOSTNAME = config.rpi_lensed_hostname
    CONFIG_FN = config.config_name
    DEFAULT_ALGO = config.default_algo

    OVERLAY_ALPHA = config.overlay.alpha
    OVERLAY_1 = config.overlay.img1
    OVERLAY_2 = config.overlay.img2
    OVERLAY_3 = config.overlay.img3

    if OVERLAY_1 is not None:
        assert os.path.exists(OVERLAY_1.fp)
    if OVERLAY_2 is not None:
        assert os.path.exists(OVERLAY_2.fp)
    if OVERLAY_3 is not None:
        assert os.path.exists(OVERLAY_3.fp)

    if config.supported_inputs is not None:
        supported_input = config.supported_inputs
    FILES_CAPTURE_CONFIG = config.files

    input_commands = ["/" + input for input in supported_input]
    HELP_TEXT = (
        "Through this bot, you can send a photo to the lensless camera setup in our lab at EPFL (shown above). "
        "The photo will be:\n\n1. Displayed on a screen.\n2. Our lensless camera will "
        "take a picture.\n3. A reconstruction will be sent back through the bot.\n4. "
        "The raw data will also be sent back."
        # "\n\nIf you do not feel comfortable sending one "
        # f"of your own pictures, you can use the {input_commands} commands to set "
        # "the image on the display with one of our inputs. Or even send an emoij 😎"
        f"\n\n⚠️ Try one of the {input_commands} commands to use images we've configured. "
        # "Or even send an emoji 😎 "
        "You can also send your own image (but brightness/exposure may need to be adjusted)."
        "\n\nAll previous data is overwritten "
        "when a new image is sent, and everything is deleted when the process running on the "
        "server is shut down."
    )

    if config.supported_algos is not None:
        supported_algos = config.supported_algos

    algo_commands = ["/" + algo for algo in supported_algos]
    ALGO_TEXT = (
        f"By default, the reconstruction is done with /{DEFAULT_ALGO}, but you "
        "can specify the algorithm (on the last measurement) with the corresponding "
        f"command: {algo_commands}."
        # "\n\nAll provided algorithms require an estimate of the point spread function (PSF). "
        # "Each user has their unique mask pattern according to the Telegram ID. "
        # "\n\n⚠️ After doing a measurement/reconstruction, you can try running /random_mask "
        # "to see what would be the reconstruction if you use a different (wrong) mask, "
        # "as if someone (like a hacker!) were trying to decode your data with a different mask."
        # "\n\nAll provided algorithms require an estimate of the point spread function (PSF). "
        # "You can measure a (proxy) PSF with /psf (a point source like "
        # "image will be displayed on the screen). "
        # "In practice, we measure the PSF with single white LED. The used PSF is sent also sent "
        # "back with the /psf command."
        "\n\nMore info: https://lensless.readthedocs.io/"
    )

    # make output folder
    if not os.path.exists(OUTPUT_FOLDER):
        os.makedirs(OUTPUT_FOLDER)

    # load and downsample PSF beforehand
    if config.psf is not None:

        if "fp" in config.psf:
            from lensless.utils.io import save_image

            psf, bg = load_psf(
                config.psf.fp, downsample=config.psf.downsample, return_float=True, return_bg=True
            )
            # save to demo folder
            PSF_FP = os.path.join(OUTPUT_FOLDER, "psf.png")
            save_image(psf[0], PSF_FP)

            # save with gamma correction
            psf_gamma = psf[0] / np.max(psf[0])
            if config.gamma > 1:
                from lensless.utils.image import gamma_correction

                psf_gamma = gamma_correction(psf_gamma, gamma=config.gamma)
            save_image(psf_gamma, PSF_FP_GAMMA)

            # save background array
            BACKGROUND_FP = os.path.join(OUTPUT_FOLDER, "psf_bg.npy")
            np.save(BACKGROUND_FP, bg)
        elif "device" in config.psf:
            # programmable mask
            MASK_PARAM = config.psf

    # Create the Application and pass it your bot's token.
    assert TOKEN is not None
    application = Application.builder().token(TOKEN).build()

    if not config.idle:

        assert RPI_USERNAME is not None
        assert RPI_HOSTNAME is not None

        check_username_hostname(RPI_USERNAME, RPI_HOSTNAME)
        if RPI_LENSED_USERNAME is not None or RPI_LENSED_HOSTNAME is not None:
            check_username_hostname(RPI_LENSED_USERNAME, RPI_LENSED_HOSTNAME)

        # on different commands - answer in Telegram
        application.add_handler(CommandHandler("start", start))
        application.add_handler(CommandHandler("help", help_command))
        application.add_handler(CommandHandler("notbusy", remove_busy_flag))

        for file_input in supported_input:
            assert file_input in FILES_CAPTURE_CONFIG.keys()
            application.add_handler(CommandHandler(file_input, file_input_command, block=False))

        # different algorithms
        application.add_handler(CommandHandler("fista", fista, block=False))
        application.add_handler(CommandHandler("admm", admm, block=False))
        application.add_handler(CommandHandler("unrolled", unrolled, block=False))
        application.add_handler(CommandHandler("unet", unet, block=False))

        # photo input
        application.add_handler(
            MessageHandler(filters.PHOTO & ~filters.COMMAND, photo, block=False)
        )

        # brightness input
        application.add_handler(CommandHandler("brightness", brightness_command, block=False))
        application.add_handler(CallbackQueryHandler(button))

        # exposure input
        application.add_handler(CommandHandler("exposure", exposure_command, block=False))
        application.add_handler(CallbackQueryHandler(button))

        # emoji input
        application.add_handler(MessageHandler(filters.TEXT & ~filters.COMMAND, emoji, block=False))

        if MASK_PARAM is not None:
            application.add_handler(CommandHandler("random_mask", random_mask, block=False))
        else:
            # to dim for measuring PSF of DigiCam?
            application.add_handler(CommandHandler("psf", psf_command, block=False))

        # Run the bot until the user presses Ctrl-C
        application.run_polling()

    else:

        application.add_handler(CommandHandler("help", help_command))
        application.add_handler(MessageHandler(None, not_running_command))

        # Run the bot until the user presses Ctrl-C
        application.run_polling()

    # delete non-empty folder
    if os.path.exists(OUTPUT_FOLDER):
        shutil.rmtree(OUTPUT_FOLDER)


if __name__ == "__main__":
    main()



import pytest
from lensless.utils.io import load_data
import numpy as np

try:
    import pycsou
    from lensless import GradientDescent, NesterovGradientDescent, FISTA, ADMM, APGD, APGDPriors

    pycsou_available = True

except ImportError:
    from lensless import GradientDescent, NesterovGradientDescent, FISTA, ADMM

    pycsou_available = False


try:
    import torch
    from lensless import UnrolledFISTA, UnrolledADMM

    torch_is_available = True
    torch.autograd.set_detect_anomaly(True)
    trainable_algos = [UnrolledFISTA, UnrolledADMM]
except ImportError:
    torch_is_available = False
    trainable_algos = []


psf_fp = "data/psf/tape_rgb.png"
data_fp = "data/raw_data/thumbs_up_rgb.png"
dtypes = ["float32", "float64"]
downsample = 16
_n_iter = 5

# classical algorithms
standard_algos = [GradientDescent, NesterovGradientDescent, FISTA, ADMM]


@pytest.mark.parametrize("algorithm", standard_algos)
def test_set_initial_est(algorithm):
    for gray in [True, False]:
        psf, _ = load_data(
            psf_fp=psf_fp,
            data_fp=data_fp,
            downsample=downsample,
            plot=False,
            gray=gray,
            use_torch=False,
        )
        recon = algorithm(psf)
        assert recon._initial_est is None
        random_init = np.random.rand(*recon._image_est_shape)
        recon._set_initial_estimate(random_init)
        assert isinstance(recon._initial_est, np.ndarray)
        assert np.allclose(recon._initial_est, random_init)

        # # set from constructor
        recon = algorithm(psf, initial_est=random_init)
        assert isinstance(recon._initial_est, np.ndarray)
        assert np.allclose(recon._initial_est, random_init)


@pytest.mark.parametrize("algorithm", trainable_algos)
def test_set_initial_est_unrolled(algorithm):
    if not torch_is_available:
        return
    for gray in [True, False]:
        psf, _ = load_data(
            psf_fp=psf_fp,
            data_fp=data_fp,
            downsample=downsample,
            plot=False,
            gray=gray,
            use_torch=True,
        )
        recon = algorithm(psf)
        assert recon._initial_est is None
        random_init = torch.rand(*recon._image_est.shape)
        recon._set_initial_estimate(random_init)
        assert isinstance(recon._initial_est, torch.Tensor)
        assert np.allclose(recon._initial_est, random_init)

        # set from constructor
        recon = algorithm(psf, initial_est=random_init)
        assert isinstance(recon._initial_est, torch.Tensor)
        assert np.allclose(recon._initial_est, random_init)


@pytest.mark.parametrize("algorithm", standard_algos)
def test_recon_numpy(algorithm):
    for gray in [True, False]:
        for dtype in dtypes:
            psf, data = load_data(
                psf_fp=psf_fp,
                data_fp=data_fp,
                downsample=downsample,
                plot=False,
                gray=gray,
                dtype=dtype,
                use_torch=False,
            )
            recon = algorithm(psf, dtype=dtype)
            recon.set_data(data)
            res = recon.apply(n_iter=_n_iter, disp_iter=None, plot=False)
            assert len(psf.shape) == 4
            assert psf.shape[3] == (1 if gray else 3)
            assert res.dtype == psf.dtype, f"Got {res.dtype}, expected {dtype}"


@pytest.mark.parametrize("algorithm", standard_algos + trainable_algos)
def test_recon_torch(algorithm):
    if not torch_is_available:
        return
    for gray in [True, False]:
        for dtype in dtypes:
            psf, data = load_data(
                psf_fp=psf_fp,
                data_fp=data_fp,
                downsample=downsample,
                plot=False,
                gray=gray,
                dtype=dtype,
                use_torch=True,
            )
            recon = algorithm(psf, dtype=dtype, n_iter=_n_iter)
            recon.set_data(data)
            res = recon.apply(disp_iter=None, plot=False)
            assert recon._n_iter == _n_iter
            assert len(psf.shape) == 4
            assert psf.shape[3] == (1 if gray else 3)
            assert res.dtype == psf.dtype, f"Got {res.dtype}, expected {dtype}"


def test_apgd():
    if pycsou_available:
        for gray in [True, False]:
            for dtype in dtypes:
                for diff_penalty in [APGDPriors.L2, None]:
                    for prox_penalty in [APGDPriors.NONNEG, APGDPriors.L1]:
                        psf, data = load_data(
                            psf_fp=psf_fp,
                            data_fp=data_fp,
                            downsample=downsample,
                            plot=False,
                            gray=gray,
                            dtype=dtype,
                            use_torch=False,
                        )
                        recon = APGD(
                            psf,
                            dtype=dtype,
                            prox_penalty=prox_penalty,
                            diff_penalty=diff_penalty,
                            rel_error=None,
                        )
                        recon.set_data(data)
                        res = recon.apply(n_iter=_n_iter, disp_iter=None, plot=False)
                        assert psf.shape[3] == (1 if gray else 3)
                        assert res.dtype == psf.dtype, f"Got {res.dtype}, expected {dtype}"
    else:
        print("Pycsou not installed. Skipping APGD test.")


#  trainable algorithms
@pytest.mark.parametrize("algorithm", trainable_algos)
def test_trainable_recon(algorithm):
    if torch_is_available:
        for dtype, torch_type in [("float32", torch.float32), ("float64", torch.float64)]:
            psf = torch.rand(1, 32, 64, 3, dtype=torch_type)
            data = torch.rand(2, 1, 32, 64, 3, dtype=torch_type)

            def pre_process(x, **kwargs):
                return x

            def post_process(x, **kwargs):
                return x

            recon = algorithm(
                psf, n_iter=_n_iter, dtype=dtype, pre_process=pre_process, post_process=post_process
            )

            assert (
                next(recon.parameters(), None) is not None
            ), f"{algorithm.__name__} has no trainable parameters"

            res = recon.forward(data)
            loss = torch.mean(res)
            loss.backward()

            assert (
                data.shape[0] == res.shape[0]
            ), f"Batch dimension changed: got {res.shape[0]} expected {data.shape[0]}"

            assert len(psf.shape) == 4
            assert res.shape[4] == 3, "Input in HWC format but output CHW format"


@pytest.mark.parametrize("algorithm", trainable_algos)
def test_trainable_batch(algorithm):
    # test if batch_call and pally give the same result for any batch size
    if not torch_is_available:
        return
    for dtype, torch_type in [("float32", torch.float32), ("float64", torch.float64)]:
        psf = torch.rand(1, 34, 64, 3, dtype=torch_type)
        data1 = torch.rand(5, 1, 34, 64, 3, dtype=torch_type)
        data2 = torch.rand(1, 1, 34, 64, 3, dtype=torch_type)
        data2[0, 0, ...] = data1[0, 0, ...]

        def pre_process(x, **kwargs):
            return x

        def post_process(x, **kwargs):
            return x

        recon = algorithm(
            psf, dtype=dtype, n_iter=_n_iter, pre_process=pre_process, post_process=post_process
        )
        res1 = recon.forward(data1)
        res2 = recon.forward(data2)
        recon.set_data(data2[0])
        res3 = recon.apply(disp_iter=None, plot=False)

        # main test
        torch.testing.assert_close(res1[0, 0, ...], res2[0, 0, ...])
        torch.testing.assert_close(res1[0, 0, ...], res3[0, ...])
        # small test
        assert res1.dtype == psf.dtype, f"Got {res1.dtype}, expected {dtype}"
        assert recon._n_iter == _n_iter
        assert len(psf.shape) == 4



"""Modified from https://github.com/Waller-Lab/DiffuserCam-Tutorial/blob/master/GD.py"""

import numpy as np
import matplotlib.pyplot as plt
from lensless.utils.plot import plot_image
import inspect
import pathlib as plib


class GradientDescentUpdate:
    """Gradient descent update techniques."""

    VANILLA = "vanilla"
    NESTEROV = "nesterov"
    FISTA = "fista"

    @staticmethod
    def all_values():
        vals = []
        for i in inspect.getmembers(GradientDescentUpdate):
            # remove private and protected functions, and this function
            if not i[0].startswith("_") and not callable(i[1]):
                vals.append(i[1])
        return vals


def create_init_matrices(psf):
    """
    Create initialization matrices

    Parameters
    ----------
    psf :py:class:`~numpy.ndarray`
        2-D array of PSF.

    Returns
    -------

    """
    pixel_start = (np.max(psf) + np.min(psf)) / 2
    x = np.ones(psf.shape) * pixel_start

    init_shape = psf.shape
    padded_shape = [next_pow2(2 * n - 1) for n in init_shape]
    starti = (padded_shape[0] - init_shape[0]) // 2
    endi = starti + init_shape[0]
    startj = (padded_shape[1] // 2) - (init_shape[1] // 2)
    endj = startj + init_shape[1]
    hpad = np.zeros(padded_shape)
    hpad[starti:endi, startj:endj] = psf

    H = np.fft.fft2(hpad, norm="ortho")
    Hadj = np.conj(H)

    def crop(X):
        return X[starti:endi, startj:endj]

    def pad(v):
        vpad = np.zeros(padded_shape).astype(np.complex64)
        vpad[starti:endi, startj:endj] = v
        return vpad

    utils = [crop, pad]
    v = np.real(pad(x))

    return H, Hadj, v, utils


def next_pow2(n):
    return int(2 ** np.ceil(np.log2(n)))


def grad(Hadj, H, vk, b, crop, pad):
    Av = calcA(H, vk, crop)
    diff = Av - b
    return np.real(calcAHerm(Hadj, diff, pad))


def calcA(H, vk, crop):
    Vk = np.fft.fft2(vk, norm="ortho")
    return crop(np.fft.ifftshift(np.fft.ifft2(H * Vk, norm="ortho")))


def calcAHerm(Hadj, diff, pad):
    xpad = pad(diff)
    X = np.fft.fft2(xpad, norm="ortho")
    return np.fft.ifftshift(np.fft.ifft2(Hadj * X, norm="ortho"))


def gd_update(params, parent_var):
    # extract variables
    vk = params["vk"]
    H, Hadj, b, crop, pad, alpha, proj = parent_var

    # apply gradient step
    gradient = grad(Hadj, H, vk, b, crop, pad)
    vk -= alpha * gradient
    vk = proj(vk)

    # update variables
    params["vk"] = vk
    return params


def nesterov_update(params, parent_var):
    # extract variables
    vk = params["vk"]
    p = params["p"]
    mu = params["mu"]
    H, Hadj, b, crop, pad, alpha, proj = parent_var

    # apply gradient step
    p_prev = p
    gradient = grad(Hadj, H, vk, b, crop, pad)
    p = mu * p - alpha * gradient
    vk += -mu * p_prev + (1 + mu) * p
    vk = proj(vk)

    # update variables
    params["vk"] = vk
    params["p"] = p
    return params


def fista_update(params, parent_var):
    # extract variables
    vk = params["vk"]
    xk = params["xk"]
    tk = params["tk"]
    H, Hadj, b, crop, pad, alpha, proj = parent_var

    # apply gradient step
    x_k1 = xk
    gradient = grad(Hadj, H, vk, b, crop, pad)
    vk -= alpha * gradient
    xk = proj(vk)
    t_k1 = (1 + np.sqrt(1 + 4 * tk**2)) / 2
    vk = xk + (tk - 1) / t_k1 * (xk - x_k1)
    tk = t_k1

    # update variables
    params["vk"] = vk
    params["tk"] = tk
    params["xk"] = xk
    return params


def non_neg(xi):
    xi = np.maximum(xi, 0)
    return xi


def grad_descent(
    psf,
    data,
    n_iter=100,
    non_neg_constraint=True,
    update_method=GradientDescentUpdate.FISTA,
    disp_iter=10,
    plot_pause=0.2,
    save=False,
):
    """
    Gradient descent reconstruction approach.

    Parameters
    ----------
    psf :py:class:`~numpy.ndarray`
        2-D array of PSF.
    data :py:class:`~numpy.ndarray`
        2-D array of raw measurement data.
    n_iter : int, optional
        Number of iterations.
    non_neg_constraint : bool, optional
        Whether to apply non-negativity constraint.
    update_method : GradientDescentUpdate member
        Which gradient descent method to apply: vanilla, Nesterov, or FISTA.
    disp_iter : int, optional
        How many iterations to wait before plotting. Set to None for no intermediate plots.
    plot_pause : float, optional
        How much to pause (in seconds) between iterations plot.

    Returns
    -------
    final_im :py:class:`~numpy.ndarray`
        2-D array of reconstructed data.
    ax : :py:class:`~matplotlib.axes.Axes`
    """
    assert (
        update_method in GradientDescentUpdate.all_values()
    ), f"update method '{update_method}' not supported."

    assert len(psf.shape) == len(data.shape)
    is_rgb = True if len(psf.shape) == 3 else False
    if not is_rgb:
        psf = psf[:, :, np.newaxis]
        data = data[:, :, np.newaxis]

    if non_neg_constraint:
        proj = non_neg  # Enforce non-negativity at every gradient step.
    else:

        def proj(x):
            return x

    H = []
    Hadj = []
    v = []
    crop = []
    pad = []
    alpha = []
    params = []
    for c in range(psf.shape[2]):
        # precompute some matrices
        _H, _Hadj, _v, _utils = create_init_matrices(psf[:, :, c])
        H.append(_H)
        Hadj.append(_Hadj)
        v.append(_v)
        crop.append(_utils[0])
        pad.append(_utils[1])

        alpha.append(np.real(1.8 / (np.max(Hadj[c] * H[c]))))

        # gradient descent variables
        _params = {"vk": _v}
        if update_method is GradientDescentUpdate.NESTEROV:
            _params.update(
                {
                    "p": 0,
                    "mu": 0.9,
                }
            )
        elif update_method is GradientDescentUpdate.FISTA:
            _params.update(
                {
                    "tk": 1,
                    "xk": _v,
                }
            )
        params.append(_params)

    if disp_iter is not None:
        ax = plot_image(data)
    else:
        ax = None
        disp_iter = n_iter + 1

    for i in range(n_iter):
        for c in range(psf.shape[2]):
            parent_var = [H[c], Hadj[c], data[:, :, c], crop[c], pad[c], alpha[c], proj]
            if update_method is GradientDescentUpdate.VANILLA:
                params[c] = gd_update(params[c], parent_var)
            elif update_method is GradientDescentUpdate.NESTEROV:
                params[c] = nesterov_update(params[c], parent_var)
            elif update_method is GradientDescentUpdate.FISTA:
                params[c] = fista_update(params[c], parent_var)

            if (i + 1) % disp_iter == 0:
                if is_rgb:
                    image = np.zeros(psf.shape)
                    for c in range(3):
                        image[:, :, c] = proj(crop[c](params[c]["vk"]))
                else:
                    image = proj(crop[0](params[0]["vk"]))
                ax = plot_image(image, ax=ax)
                ax.set_title("Reconstruction after iteration {}".format(i + 1))
                if save:
                    plt.savefig(plib.Path(save) / f"{i + 1}_diffusercam.png")
                plt.draw()
                plt.pause(plot_pause)

    if is_rgb:
        final_im = np.zeros(psf.shape)
        for c in range(3):
            final_im[:, :, c] = proj(crop[c](params[c]["vk"]))
    else:
        final_im = proj(crop[0](params[0]["vk"]))
    ax = plot_image(final_im, ax=ax)
    ax.set_title("Final reconstruction after {} iterations".format(n_iter))
    if save:
        plt.savefig(plib.Path(save) / f"{n_iter}_diffusercam.png")

    return final_im, ax



import numpy as np
from lensless.hardware.mask import CodedAperture, PhaseContour, FresnelZoneAperture
from lensless.eval.metric import mse, psnr, ssim
from waveprop.fresnel import fresnel_conv


resolution = np.array([380, 507])
d1 = 3e-6
dz = 4e-3


def test_flatcam():

    mask1 = CodedAperture(
        method="MURA",
        n_bits=23,
        resolution=resolution,
        feature_size=d1,
        distance_sensor=dz,
    )
    assert np.all(mask1.mask.shape == resolution)

    desired_psf_shape = np.array(tuple(resolution) + (len(mask1.psf_wavelength),))
    assert np.all(mask1.psf.shape == desired_psf_shape)

    mask2 = CodedAperture(
        method="MLS",
        n_bits=5,
        resolution=resolution,
        feature_size=d1,
        distance_sensor=dz,
    )
    assert np.all(mask2.mask.shape == resolution)
    desired_psf_shape = np.array(tuple(resolution) + (len(mask2.psf_wavelength),))
    assert np.all(mask2.psf.shape == desired_psf_shape)


def test_phlatcam():

    mask = PhaseContour(
        noise_period=(8, 8),
        resolution=resolution,
        feature_size=d1,
        distance_sensor=dz,
    )
    assert np.all(mask.shape == resolution)
    desired_psf_shape = np.array(tuple(resolution) + (len(mask.psf_wavelength),))
    assert np.all(mask.psf.shape == desired_psf_shape)

    u1 = mask.height_map_to_field(wavelength=mask.design_wv)
    Mp = np.sqrt(mask.target_psf) * np.exp(
        1j * np.angle(fresnel_conv(u1, mask.design_wv, d1, dz, dtype=np.float32)[0])
    )
    assert mse(abs(Mp), np.sqrt(mask.target_psf)) < 0.1
    assert psnr(abs(Mp), np.sqrt(mask.target_psf)) > 30
    assert abs(1 - ssim(abs(Mp), np.sqrt(mask.target_psf), channel_axis=None)) < 0.1


def test_fza():

    mask = FresnelZoneAperture(
        radius=30.0, resolution=resolution, feature_size=d1, distance_sensor=dz
    )
    assert np.all(mask.mask.shape == resolution)
    desired_psf_shape = np.array(tuple(resolution) + (len(mask.psf_wavelength),))
    assert np.all(mask.psf.shape == desired_psf_shape)


def test_classmethod():

    downsample = 8

    mask1 = CodedAperture.from_sensor(
        sensor_name="rpi_hq", downsample=downsample, distance_sensor=dz
    )
    assert np.all(mask1.shape == resolution)
    desired_psf_shape = np.array(tuple(resolution) + (len(mask1.psf_wavelength),))
    assert np.all(mask1.psf.shape == desired_psf_shape)

    mask2 = PhaseContour.from_sensor(
        sensor_name="rpi_hq", downsample=downsample, distance_sensor=dz
    )
    assert np.all(mask2.shape == resolution)
    desired_psf_shape = np.array(tuple(resolution) + (len(mask2.psf_wavelength),))
    assert np.all(mask2.psf.shape == desired_psf_shape)

    mask3 = FresnelZoneAperture.from_sensor(
        sensor_name="rpi_hq", downsample=downsample, distance_sensor=dz
    )
    assert np.all(mask3.shape == resolution)
    desired_psf_shape = np.array(tuple(resolution) + (len(mask3.psf_wavelength),))
    assert np.all(mask3.psf.shape == desired_psf_shape)


if __name__ == "__main__":
    test_flatcam()
    test_phlatcam()
    test_fza()
    test_classmethod()



"""
Check maximum pixel value of images and check for saturation / underexposure.

```
python scripts/measure/analyze_measured_dataset.py dataset_path=PATH
```
"""

import hydra
from hydra.utils import to_absolute_path
import glob
import os
from PIL import Image
import numpy as np
import matplotlib.pyplot as plt
import time
import tqdm
import re


def convert(text):
    return int(text) if text.isdigit() else text.lower()


def alphanum_key(key):
    return [convert(c) for c in re.split("([0-9]+)", key)]


def natural_sort(arr):
    return sorted(arr, key=alphanum_key)


@hydra.main(version_base=None, config_path="../../configs", config_name="analyze_dataset")
def analyze_dataset(config):

    folder = config.dataset_path
    desired_range = config.desired_range
    delete_bad = config.delete_bad
    start_idx = config.start_idx
    saturation_percent = config.saturation_percent

    assert (
        folder is not None
    ), "Must specify folder to analyze in config or through command line (folder=PATH)."

    # get all PNG files in folder
    files = natural_sort(glob.glob(os.path.join(folder, "*.png")))
    print("Found {} files".format(len(files)))
    if start_idx is not None:
        files = files[start_idx:]
        print("Starting at file {}".format(files[0]))
    if config.n_files is not None:
        files = files[: config.n_files]
        print("Analyzing first {} files".format(len(files)))

    # loop over files for maximum value
    max_vals = []
    n_bad_files = 0
    bad_files = []
    start_time = time.time()
    for fn in tqdm.tqdm(files):
        im = np.array(Image.open(fn))
        max_val = im.max()
        max_vals.append(max_val)
        saturation_ratio = np.sum(im >= desired_range[1]) / im.size

        if max_val < desired_range[0]:
            n_bad_files += 1
            bad_files.append(fn)

            if delete_bad:
                os.remove(fn)
                print("REMOVED file {}".format(fn))
            else:
                print("File {} has max value {}".format(fn, max_val))

        elif saturation_ratio > saturation_percent:
            n_bad_files += 1
            bad_files.append(fn)

            if delete_bad:
                os.remove(fn)
                print("REMOVED file {}".format(fn))
            else:
                print("File {} has saturation ratio {}".format(fn, saturation_ratio))

        # # if out of desired range, print filename
        # if max_val < desired_range[0] or saturation_ratio > saturation_percent:
        #     # print("File {} has max value {}".format(fn, max_val))
        #     n_bad_files += 1
        #     bad_files.append(fn)

        #     if delete_bad:
        #         os.remove(fn)
        #         print("REMOVED file {}".format(fn))
        #     else:
        #         print("File {} has max value {}".format(fn, max_val))

    proc_time = time.time() - start_time
    print("Went through {} files in {:.2f} seconds".format(len(files), proc_time))
    print(
        "Found {} / {} bad files  ({}%)".format(
            n_bad_files, len(files), n_bad_files / len(files) * 100
        )
    )

    # plot histogram
    output_folder = os.getcwd()
    output_fp = os.path.join(output_folder, "max_vals.png")
    plt.hist(max_vals, bins=100)
    plt.savefig(output_fp)

    print("Saved histogram to {}".format(output_fp))

    # command line input on whether to delete bad files
    if not delete_bad:
        response = None
        while response not in ["yes", "no"]:
            response = input("Delete bad files: [yes|no] : ")
        if response == "yes":
            for _fn in bad_files:
                os.remove(_fn)
        else:
            print("Not deleting bad files")

    # check for matching background file
    files_bg = natural_sort(glob.glob(os.path.join(folder, "black_background*.png")))
    # -- remove files_bg from files
    files = [fn for fn in files if fn not in files_bg]

    if len(files_bg) > 0:
        print("Found {} background files".format(len(files_bg)))
        # detect files that don't have background
        files_no_bg = []
        for fn in files:
            bn = os.path.basename(fn).split(".")[0]
            _bg_file = os.path.join(folder, "black_background{}.png".format(bn))
            if _bg_file not in files_bg:
                files_no_bg.append(fn)

        print("Found {} files without background".format(len(files_no_bg)))
        # ask to delete files without background
        response = None
        while response not in ["yes", "no"]:
            response = input("Delete files without background: [yes|no] : ")
        if response == "yes":
            for _fn in files_no_bg:
                if os.path.exists(_fn):  # maybe already deleted before
                    os.remove(_fn)


if __name__ == "__main__":
    analyze_dataset()



# #############################################################################
# unrolled_admm.py
# =================
# Authors :
# Yohann PERRON [yohann.perron@gmail.com]
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

from lensless.recon.trainable_recon import TrainableReconstructionAlgorithm
from lensless.recon.admm import soft_thresh, finite_diff, finite_diff_adj, finite_diff_gram


try:
    import torch

except ImportError:
    raise ImportError("Pytorch is require to use trainable reconstruction algorithms.")


class UnrolledADMM(TrainableReconstructionAlgorithm):
    """
    Object for applying unrolled version of :py:class:`~lensless.ADMM`.

    """

    def __init__(
        self,
        psf,
        dtype=None,
        n_iter=5,
        mu1=1e-6,
        mu2=1e-5,
        mu3=4e-5,
        tau=0.0001,
        psi=None,
        psi_adj=None,
        psi_gram=None,
        pad=False,
        norm="backward",
        **kwargs,
    ):
        """

        Parameters
        ----------
        psf : :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        dtype : float32 or float64
            Data type to use for optimization. Default is float32.
        n_iter : int, optional
            Number of iterations to unrolled, by default 5
        mu1 : float
            Initial step size for updating primal/dual variables.
        mu2 : float
            Initial step size for updating primal/dual variables.
        mu3 : float
            Initial step size for updating primal/dual variables.
        tau : float
            Initial weight for L1 norm of `psi` applied to the image estimate.
        psi : :py:class:`function`, optional
            Operator to map image to a space that the image is assumed to be
            sparse in (hence L1 norm). Default is to use total variation (TV)
            operator.
        psi_adj : :py:class:`function`
            Adjoint of `psi`.
        psi_gram : :py:class:`function`
            Function to compute gram of `psi`.
        pad : bool, optional
            Whether to pad the image with zeros before applying the PSF.
        norm : str
            Normalization to use for the convolution. Options are "forward",
            "backward", and "ortho". Default is "backward".
        """

        super(UnrolledADMM, self).__init__(
            psf, n_iter=n_iter, dtype=dtype, pad=pad, norm=norm, reset=False, **kwargs
        )

        if not self.skip_unrolled:
            self._mu1_p = torch.nn.Parameter(
                torch.ones(self._n_iter, device=self._psf.device) * mu1
            )
            self._mu2_p = torch.nn.Parameter(
                torch.ones(self._n_iter, device=self._psf.device) * mu2
            )
            self._mu3_p = torch.nn.Parameter(
                torch.ones(self._n_iter, device=self._psf.device) * mu3
            )
            self._tau_p = torch.nn.Parameter(
                torch.ones(self._n_iter, device=self._psf.device) * tau
            )
        else:
            self._mu1_p = torch.ones(self._n_iter, device=self._psf.device) * mu1
            self._mu2_p = torch.ones(self._n_iter, device=self._psf.device) * mu2
            self._mu3_p = torch.ones(self._n_iter, device=self._psf.device) * mu3
            self._tau_p = torch.ones(self._n_iter, device=self._psf.device) * tau

        # set prior
        if psi is None:
            # use already defined Psi and PsiT
            self._PsiTPsi = finite_diff_gram(self._padded_shape, self._dtype, self.is_torch)
        else:
            assert psi_adj is not None
            assert psi_gram is not None
            assert callable(psi)
            assert callable(psi_adj)
            assert callable(psi_gram)
            # overwrite already defined Psi and PsiT
            self._Psi = psi
            self._PsiT = psi_adj
            self._PsiTPsi = psi_gram(self._padded_shape)

        self._PsiTPsi = self._PsiTPsi.to(self._psf.device)

        self.reset()

    def _Psi(self, x):
        """
        Operator to map image to space that the image is assumed to be sparse
        in.
        """
        return finite_diff(x)

    def _PsiT(self, U):
        """
        Adjoint of `_Psi`.
        """
        return finite_diff_adj(U)

    def reset(self, batch_size=1):

        if self._data is not None:
            device = self._data.device
        else:
            device = self._convolver._H.device

        # ensure that mu1, mu2, mu3, tau are positive
        self._mu1 = torch.abs(self._mu1_p).to(device)
        self._mu2 = torch.abs(self._mu2_p).to(device)
        self._mu3 = torch.abs(self._mu3_p).to(device)
        self._tau = torch.abs(self._tau_p).to(device)

        # TODO initialize without padding
        if self._initial_est is not None:
            self._image_est = self._initial_est.to(device)
        else:
            self._image_est = torch.zeros([1] + self._padded_shape, dtype=self._dtype).to(device)

        self._X = torch.zeros_like(self._image_est)
        self._U = torch.zeros_like(self._Psi(self._image_est))
        self._W = torch.zeros_like(self._X)
        if self._image_est.max():
            # if non-zero
            self._forward_out = self._convolver.convolve(self._image_est)
            self._Psi_out = self._Psi(self._image_est)
        else:
            self._forward_out = torch.zeros_like(self._X)
            self._Psi_out = torch.zeros_like(self._U)

        self._xi = torch.zeros_like(self._image_est)
        self._eta = torch.zeros_like(self._U)
        self._rho = torch.zeros_like(self._X)

        # precompute_R_divmat [iter, batch, depth, height, width, channels]
        self._R_divmat = 1.0 / (
            self._mu1[:, None, None, None, None, None]
            * (torch.abs(self._convolver._Hadj * self._convolver._H))[None, ...]
            + self._mu2[:, None, None, None, None, None] * torch.abs(self._PsiTPsi).to(device)
            + self._mu3[:, None, None, None, None, None]
        ).type(self._complex_dtype)

        # precompute_X_divmat [iter, batch, depth, height, width, channels]
        self._X_divmat = 1.0 / (
            self._convolver._pad(torch.ones_like(self._convolver._psf))[None, ...]
            + self._mu1[:, None, None, None, None, None]
        )

    def _U_update(self, iter):
        """Total variation update."""
        # to avoid computing sparse operator twice
        self._U = soft_thresh(
            self._Psi_out + self._eta / self._mu2[iter], self._tau[iter] / self._mu2[iter]
        )

    def _X_update(self, iter):
        # to avoid computing forward model twice
        self._X = self._X_divmat[iter] * (
            self._xi + self._mu1[iter] * self._forward_out + self._convolver._pad(self._data)
        )

    def _image_update(self, iter):
        rk = (
            (self._mu3[iter] * self._W - self._rho)
            + self._PsiT(self._mu2[iter] * self._U - self._eta)
            + self._convolver.deconvolve(self._mu1[iter] * self._X - self._xi)
        )
        freq_space_result = self._R_divmat[iter] * torch.fft.rfft2(rk, dim=(-3, -2))
        self._image_est = torch.fft.irfft2(
            freq_space_result, dim=(-3, -2), s=self._convolver._padded_shape[-3:-1]
        )

    def _W_update(self, iter):
        """Non-negativity update"""
        self._W = torch.maximum(
            self._rho / self._mu3[iter] + self._image_est, torch.zeros_like(self._image_est)
        )

    def _xi_update(self, iter):
        # to avoid computing forward model twice
        self._xi = self._xi + self._mu1[iter] * (self._forward_out - self._X)

    def _eta_update(self, iter):
        # to avoid finite difference operataion again?
        self._eta = self._eta + self._mu2[iter] * (self._Psi_out - self._U)

    def _rho_update(self, iter):
        self._rho = self._rho + self._mu3[iter] * (self._image_est - self._W)

    def _update(self, iter):
        self._U_update(iter)
        self._X_update(iter)
        self._W_update(iter)
        self._image_update(iter)

        # update forward and sparse operators
        self._forward_out = self._convolver.convolve(self._image_est)
        self._Psi_out = self._Psi(self._image_est)

        self._xi_update(iter)
        self._eta_update(iter)
        self._rho_update(iter)

    def _form_image(self):
        image = self._convolver._crop(self._image_est)
        # image = torch.clamp(image, min=0)
        image = torch.clip(image, min=0.0)
        return image



# #############################################################################
# basicblock.py
# =================
# Original Repo:
# https://github.com/cszn/DPIR/blob/15bca3fcc1f3cc51a1f99ccf027691e278c19354/models/basicblock.py
# #############################################################################

from collections import OrderedDict
import torch.nn as nn


"""
# --------------------------------------------
# Advanced nn.Sequential
# https://github.com/xinntao/BasicSR
# --------------------------------------------
"""


def sequential(*args):
    """Advanced nn.Sequential.

    Args:
        nn.Sequential, nn.Module

    Returns:
        nn.Sequential
    """
    if len(args) == 1:
        if isinstance(args[0], OrderedDict):
            raise NotImplementedError("sequential does not support OrderedDict input.")
        return args[0]  # No sequential is needed.
    modules = []
    for module in args:
        if isinstance(module, nn.Sequential):
            for submodule in module.children():
                modules.append(submodule)
        elif isinstance(module, nn.Module):
            modules.append(module)
    return nn.Sequential(*modules)


"""
# --------------------------------------------
# Useful blocks
# https://github.com/xinntao/BasicSR
# --------------------------------
# conv + normaliation + relu (conv)
# (PixelUnShuffle)
# (ConditionalBatchNorm2d)
# concat (ConcatBlock)
# sum (ShortcutBlock)
# resblock (ResBlock)
# Channel Attention (CA) Layer (CALayer)
# Residual Channel Attention Block (RCABlock)
# Residual Channel Attention Group (RCAGroup)
# Residual Dense Block (ResidualDenseBlock_5C)
# Residual in Residual Dense Block (RRDB)
# --------------------------------------------
"""


# --------------------------------------------
# return nn.Sequantial of (Conv + BN + ReLU)
# --------------------------------------------
def conv(
    in_channels=64,
    out_channels=64,
    kernel_size=3,
    stride=1,
    padding=1,
    bias=True,
    mode="CBR",
    negative_slope=0.2,
):
    L = []
    for t in mode:
        if t == "C":
            L.append(
                nn.Conv2d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=kernel_size,
                    stride=stride,
                    padding=padding,
                    bias=bias,
                )
            )
        elif t == "T":
            L.append(
                nn.ConvTranspose2d(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    kernel_size=kernel_size,
                    stride=stride,
                    padding=padding,
                    bias=bias,
                )
            )
        elif t == "B":
            L.append(nn.BatchNorm2d(out_channels, momentum=0.9, eps=1e-04, affine=True))
        elif t == "I":
            L.append(nn.InstanceNorm2d(out_channels, affine=True))
        elif t == "R":
            L.append(nn.ReLU(inplace=True))
        elif t == "r":
            L.append(nn.ReLU(inplace=False))
        elif t == "L":
            L.append(nn.LeakyReLU(negative_slope=negative_slope, inplace=True))
        elif t == "l":
            L.append(nn.LeakyReLU(negative_slope=negative_slope, inplace=False))
        elif t == "2":
            L.append(nn.PixelShuffle(upscale_factor=2))
        elif t == "3":
            L.append(nn.PixelShuffle(upscale_factor=3))
        elif t == "4":
            L.append(nn.PixelShuffle(upscale_factor=4))
        elif t == "U":
            L.append(nn.Upsample(scale_factor=2, mode="nearest"))
        elif t == "u":
            L.append(nn.Upsample(scale_factor=3, mode="nearest"))
        elif t == "v":
            L.append(nn.Upsample(scale_factor=4, mode="nearest"))
        elif t == "M":
            L.append(nn.MaxPool2d(kernel_size=kernel_size, stride=stride, padding=0))
        elif t == "A":
            L.append(nn.AvgPool2d(kernel_size=kernel_size, stride=stride, padding=0))
        else:
            raise NotImplementedError("Undefined type: {}".format(t))
    return sequential(*L)


# --------------------------------------------
# Res Block: x + conv(relu(conv(x)))
# --------------------------------------------
class ResBlock(nn.Module):
    def __init__(
        self,
        in_channels=64,
        out_channels=64,
        kernel_size=3,
        stride=1,
        padding=1,
        bias=True,
        mode="CRC",
        negative_slope=0.2,
    ):
        super(ResBlock, self).__init__()

        assert in_channels == out_channels, "Only support in_channels==out_channels."
        if mode[0] in ["R", "L"]:
            mode = mode[0].lower() + mode[1:]

        self.res = conv(
            in_channels, out_channels, kernel_size, stride, padding, bias, mode, negative_slope
        )

    def forward(self, x):
        return x + self.res(x)


"""
# --------------------------------------------
# Upsampler
# Kai Zhang, https://github.com/cszn/KAIR
# --------------------------------------------
# upsample_pixelshuffle
# upsample_upconv
# upsample_convtranspose
# --------------------------------------------
"""


# --------------------------------------------
# conv + subp (+ relu)
# --------------------------------------------
def upsample_pixelshuffle(
    in_channels=64,
    out_channels=3,
    kernel_size=3,
    stride=1,
    padding=1,
    bias=True,
    mode="2R",
    negative_slope=0.2,
):
    assert len(mode) < 4 and mode[0] in ["2", "3", "4"], "mode examples: 2, 2R, 2BR, 3, ..., 4BR."
    up1 = conv(
        in_channels,
        out_channels * (int(mode[0]) ** 2),
        kernel_size,
        stride,
        padding,
        bias,
        mode="C" + mode,
        negative_slope=negative_slope,
    )
    return up1


# --------------------------------------------
# nearest_upsample + conv (+ R)
# --------------------------------------------
def upsample_upconv(
    in_channels=64,
    out_channels=3,
    kernel_size=3,
    stride=1,
    padding=1,
    bias=True,
    mode="2R",
    negative_slope=0.2,
):
    assert len(mode) < 4 and mode[0] in ["2", "3", "4"], "mode examples: 2, 2R, 2BR, 3, ..., 4BR"
    if mode[0] == "2":
        uc = "UC"
    elif mode[0] == "3":
        uc = "uC"
    elif mode[0] == "4":
        uc = "vC"
    mode = mode.replace(mode[0], uc)
    up1 = conv(
        in_channels,
        out_channels,
        kernel_size,
        stride,
        padding,
        bias,
        mode=mode,
        negative_slope=negative_slope,
    )
    return up1


# --------------------------------------------
# convTranspose (+ relu)
# --------------------------------------------
def upsample_convtranspose(
    in_channels=64,
    out_channels=3,
    kernel_size=2,
    stride=2,
    padding=0,
    bias=True,
    mode="2R",
    negative_slope=0.2,
):
    assert len(mode) < 4 and mode[0] in ["2", "3", "4"], "mode examples: 2, 2R, 2BR, 3, ..., 4BR."
    kernel_size = int(mode[0])
    stride = int(mode[0])
    mode = mode.replace(mode[0], "T")
    up1 = conv(in_channels, out_channels, kernel_size, stride, padding, bias, mode, negative_slope)
    return up1


"""
# --------------------------------------------
# Downsampler
# Kai Zhang, https://github.com/cszn/KAIR
# --------------------------------------------
# downsample_strideconv
# downsample_maxpool
# downsample_avgpool
# --------------------------------------------
"""


# --------------------------------------------
# strideconv (+ relu)
# --------------------------------------------
def downsample_strideconv(
    in_channels=64,
    out_channels=64,
    kernel_size=2,
    stride=2,
    padding=0,
    bias=True,
    mode="2R",
    negative_slope=0.2,
):
    assert len(mode) < 4 and mode[0] in ["2", "3", "4"], "mode examples: 2, 2R, 2BR, 3, ..., 4BR."
    kernel_size = int(mode[0])
    stride = int(mode[0])
    mode = mode.replace(mode[0], "C")
    down1 = conv(
        in_channels, out_channels, kernel_size, stride, padding, bias, mode, negative_slope
    )
    return down1


# --------------------------------------------
# maxpooling + conv (+ relu)
# --------------------------------------------
def downsample_maxpool(
    in_channels=64,
    out_channels=64,
    kernel_size=3,
    stride=1,
    padding=0,
    bias=True,
    mode="2R",
    negative_slope=0.2,
):
    assert len(mode) < 4 and mode[0] in ["2", "3"], "mode examples: 2, 2R, 2BR, 3, ..., 3BR."
    kernel_size_pool = int(mode[0])
    stride_pool = int(mode[0])
    mode = mode.replace(mode[0], "MC")
    pool = conv(
        kernel_size=kernel_size_pool,
        stride=stride_pool,
        mode=mode[0],
        negative_slope=negative_slope,
    )
    pool_tail = conv(
        in_channels,
        out_channels,
        kernel_size,
        stride,
        padding,
        bias,
        mode=mode[1:],
        negative_slope=negative_slope,
    )
    return sequential(pool, pool_tail)


# --------------------------------------------
# averagepooling + conv (+ relu)
# --------------------------------------------
def downsample_avgpool(
    in_channels=64,
    out_channels=64,
    kernel_size=3,
    stride=1,
    padding=1,
    bias=True,
    mode="2R",
    negative_slope=0.2,
):
    assert len(mode) < 4 and mode[0] in ["2", "3"], "mode examples: 2, 2R, 2BR, 3, ..., 3BR."
    kernel_size_pool = int(mode[0])
    stride_pool = int(mode[0])
    mode = mode.replace(mode[0], "AC")
    pool = conv(
        kernel_size=kernel_size_pool,
        stride=stride_pool,
        mode=mode[0],
        negative_slope=negative_slope,
    )
    pool_tail = conv(
        in_channels,
        out_channels,
        kernel_size,
        stride,
        padding,
        bias,
        mode=mode[1:],
        negative_slope=negative_slope,
    )
    return sequential(pool, pool_tail)



import numpy as np
import os
import time
import hydra
import torch
from hydra.utils import to_absolute_path
import matplotlib.pyplot as plt
from slm_controller import slm
from lensless.utils.io import save_image, get_dtype, load_psf
from lensless.utils.plot import plot_image
from lensless.hardware.sensor import VirtualSensor
from lensless.hardware.slm import get_programmable_mask, get_intensity_psf
from waveprop.devices import slm_dict
from waveprop.devices import SLMParam as SLMParam_wp
from huggingface_hub import hf_hub_download
from lensless.utils.image import gamma_correction


@hydra.main(version_base=None, config_path="../../configs/simulate", config_name="sim_digicam_psf")
def digicam_psf(config):

    output_folder = os.getcwd()

    # get file paths
    fp_psf = None
    if config.huggingface_repo is not None:
        # download from huggingface
        fp = hf_hub_download(
            repo_id=config.huggingface_repo,
            filename=config.huggingface_mask_pattern,
            repo_type="dataset",
        )
        if config.huggingface_psf is not None:
            fp_psf = hf_hub_download(
                repo_id=config.huggingface_repo,
                filename=config.huggingface_psf,
                repo_type="dataset",
            )
    else:
        fp = to_absolute_path(config.digicam.pattern)
        if config.digicam.psf is not None:
            fp_psf = to_absolute_path(config.digicam.psf)
    bn = os.path.basename(fp).split(".")[0]

    # digicam config
    ap_center = np.array(config.digicam.ap_center)
    ap_shape = np.array(config.digicam.ap_shape)
    rotate_angle = config.digicam.rotate
    slm_param = slm_dict[config.digicam.slm]
    sensor = VirtualSensor.from_name(
        config.digicam.sensor,
        downsample=config.digicam.downsample if config.digicam.downsample > 1 else None,
    )

    # simulation parameters
    scene2mask = config.sim.scene2mask
    mask2sensor = config.sim.mask2sensor

    torch_device = config.torch_device
    dtype = get_dtype(config.dtype, config.use_torch)

    """
    Load pattern
    """
    pattern = np.load(fp)
    # - make random pattern like original
    # pattern = np.random.rand(*pattern.shape) * 255
    # pattern = pattern.astype(np.uint8)

    # -- apply aperture
    aperture = np.zeros(pattern.shape, dtype=np.uint8)
    top_left = np.array(ap_center) - np.array(ap_shape) // 2
    bottom_right = top_left + np.array(ap_shape)
    aperture[:, top_left[0] : bottom_right[0], top_left[1] : bottom_right[1]] = 1
    pattern = pattern * aperture

    # -- extract aperture region
    idx_1 = ap_center[0] - ap_shape[0] // 2
    idx_2 = ap_center[1] - ap_shape[1] // 2

    pattern_sub = pattern[
        :,
        idx_1 : idx_1 + ap_shape[0],
        idx_2 : idx_2 + ap_shape[1],
    ]

    print("Controllable region shape: ", pattern_sub.shape)
    print("Total number of pixels: ", np.prod(pattern_sub.shape))

    # -- plot full
    s = slm.create(config.digicam.slm)
    s.set_preview(True)
    s.imshow(pattern)
    plt.savefig(os.path.join(output_folder, "pattern.png"))

    # -- plot sub pattern
    plt.imshow(pattern_sub.transpose(1, 2, 0))
    plt.savefig(os.path.join(output_folder, "pattern_sub.png"))

    """
    Simulate PSF
    """
    start_time = time.time()
    slm_vals = pattern_sub / 255.0

    # prepare color filter
    if SLMParam_wp.COLOR_FILTER in slm_param.keys():
        color_filter = slm_param[SLMParam_wp.COLOR_FILTER]
        if config.use_torch:
            color_filter = torch.from_numpy(color_filter.copy()).to(
                device=torch_device, dtype=dtype
            )
        else:
            color_filter = color_filter.astype(dtype)

    if config.digicam.slm == "adafruit":
        # flatten color channel along rows
        slm_vals = slm_vals.reshape((-1, slm_vals.shape[-1]), order="F")

    # save extracted mask values
    np.save(os.path.join(output_folder, "mask_vals.npy"), slm_vals)

    if config.use_torch:
        slm_vals = torch.from_numpy(slm_vals).to(device=torch_device, dtype=dtype)
    else:
        slm_vals = slm_vals.astype(dtype)

    # -- get mask
    mask = get_programmable_mask(
        vals=slm_vals,
        sensor=sensor,
        slm_param=slm_param,
        rotate=rotate_angle,
        flipud=config.sim.flipud,
        color_filter=color_filter,
        deadspace=config.sim.deadspace,
    )

    if config.digicam.vertical_shift is not None:
        if config.use_torch:
            mask = torch.roll(
                mask, config.digicam.vertical_shift // config.digicam.downsample, dims=1
            )
        else:
            mask = np.roll(mask, config.digicam.vertical_shift // config.digicam.downsample, axis=1)

    if config.digicam.horizontal_shift is not None:
        if config.use_torch:
            mask = torch.roll(
                mask, config.digicam.horizontal_shift // config.digicam.downsample, dims=2
            )
        else:
            mask = np.roll(
                mask, config.digicam.horizontal_shift // config.digicam.downsample, axis=2
            )

    # -- plot mask
    if config.use_torch:
        mask_np = mask.cpu().detach().numpy()
    else:
        mask_np = mask.copy()
    mask_np = np.transpose(mask_np, (1, 2, 0))
    plt.imshow(mask_np)
    plt.savefig(os.path.join(output_folder, "mask.png"))

    # -- propagate to sensor
    psf_in = get_intensity_psf(
        mask=mask,
        sensor=sensor,
        waveprop=config.sim.waveprop,
        scene2mask=scene2mask,
        mask2sensor=mask2sensor,
    )

    # -- plot PSF
    if config.use_torch:
        psf_in_np = psf_in.cpu().detach().numpy()
    else:
        psf_in_np = psf_in.copy()
    psf_in_np = np.transpose(psf_in_np, (1, 2, 0))

    # plot
    psf_meas = None
    if fp_psf is not None:
        if os.path.exists(fp_psf):
            psf_meas = load_psf(fp_psf, downsample=config.digicam.downsample)
        else:
            print("Could not load PSF image from: ", fp_psf)

    fp = os.path.join(output_folder, "sim_psf_plot.png")
    fig = plt.figure(frameon=False)
    ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
    ax.set_axis_off()
    fig.add_axes(ax)
    # ax.imshow(psf_in_np / np.max(psf_in_np))
    plot_image(psf_in_np, gamma=config.digicam.gamma, normalize=True, ax=ax)
    ax.set_xticks([])
    ax.set_yticks([])
    plt.savefig(fp)

    if psf_meas is not None:

        fig = plt.figure(frameon=False)
        ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
        ax.set_axis_off()
        fig.add_axes(ax)
        plot_image(psf_meas, gamma=config.digicam.gamma, normalize=True, ax=ax)
        # remove axis values
        ax.set_xticks([])
        ax.set_yticks([])
        plt.savefig(os.path.join(output_folder, "meas_psf_plot.png"))

        # plot overlayed
        fp = os.path.join(output_folder, "psf_overlay.png")
        psf_meas_norm = psf_meas[0] / np.max(psf_meas)
        psf_in_np_norm = psf_in_np / np.max(psf_in_np)

        plt.figure()
        plt.imshow(psf_in_np_norm, alpha=0.7)
        plt.imshow(psf_meas_norm, alpha=0.4)
        plt.savefig(fp)

    # save PSF as png
    fp = os.path.join(output_folder, f"{bn}_SIM_psf.png")
    # if config.digicam.gamma > 1:
    #     psf_in_np = psf_in_np / psf_in_np.max()
    #     psf_in_np = gamma_correction(psf_in_np, gamma=config.digicam.gamma)
    save_image(psf_in_np, fp)

    proc_time = time.time() - start_time
    print(f"\nProcessing time: {proc_time:.2f} seconds")

    print(f"\nFiles saved to : {output_folder}")


if __name__ == "__main__":
    digicam_psf()



# #############################################################################
# slm.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################


import os
import numpy as np
from lensless.hardware.utils import check_username_hostname
from lensless.utils.io import get_ctypes
from slm_controller.hardware import SLMParam, slm_devices
from scipy.ndimage import rotate as rotate_func


try:
    import torch
    from torchvision import transforms
    from torchvision.transforms.functional import InterpolationMode

    torch_available = True
except ImportError:
    torch_available = False

try:
    from waveprop.spherical import spherical_prop
    from waveprop.color import ColorSystem
    from waveprop.rs import angular_spectrum
    from waveprop.slm import get_centers
    from waveprop.devices import SLMParam as SLMParam_wp

    waveprop_available = True
except ImportError:
    waveprop_available = False


SUPPORTED_DEVICE = {
    "adafruit": "~/slm-controller/examples/adafruit_slm.py",
    "nokia": "~/slm-controller/examples/nokia_slm.py",
    "holoeye": "~/slm-controller/examples/holoeye_slm.py",
}


def set_programmable_mask(pattern, device, rpi_username=None, rpi_hostname=None, verbose=False):
    """
    Set LCD pattern on Raspberry Pi.

    This function assumes that `slm-controller <https://github.com/ebezzam/slm-controller>`_
    is installed on the Raspberry Pi.

    Parameters
    ----------
    pattern : :py:class:`~numpy.ndarray`
        Pattern to set on programmable mask.
    device : str
        Name of device to set pattern on. Supported devices: "adafruit", "nokia", "holoeye".
    rpi_username : str
        Username of Raspberry Pi.
    rpi_hostname : str
        Hostname of Raspberry Pi.

    """

    if rpi_username is not None:
        assert (
            rpi_hostname is not None
        ), "rpi_hostname must be specified if rpi_username is specified"
        client = check_username_hostname(rpi_username, rpi_hostname)

    # get path to python executable on Raspberry Pi
    rpi_python = "~/slm-controller/slm_controller_env/bin/python"
    assert (
        device in SUPPORTED_DEVICE.keys()
    ), f"Device {device} not supported. Supported devices: {SUPPORTED_DEVICE.keys()}"
    script = SUPPORTED_DEVICE[device]

    # check that pattern is correct shape
    expected_shape = slm_devices[device][SLMParam.SLM_SHAPE]
    if not slm_devices[device][SLMParam.MONOCHROME]:
        expected_shape = (3, *expected_shape)
    assert (
        pattern.shape == expected_shape
    ), f"Pattern shape {pattern.shape} does not match expected shape {expected_shape}"

    # save pattern
    pattern_fn = "tmp_pattern.npy"
    local_path = os.path.join(os.getcwd(), pattern_fn)
    np.save(local_path, pattern)

    if rpi_username is not None:

        # copy pattern to Raspberry Pi
        remote_path = f"~/{pattern_fn}"
        if verbose:
            print(f"PUTTING {local_path} to {remote_path}")
        os.system(
            'scp %s "%s@%s:%s" >/dev/null 2>&1'
            % (local_path, rpi_username, rpi_hostname, remote_path)
        )
        # # -- not sure why this doesn't work... permission denied
        # sftp = client.open_sftp()
        # sftp.put(local_path, remote_path, confirm=True)
        # sftp.close()

        # run script on Raspberry Pi to set mask pattern
        command = f"{rpi_python} {script} --file_path {remote_path}"
        if verbose:
            print(f"COMMAND : {command}")
        _stdin, _stdout, _stderr = client.exec_command(command)
        if verbose:
            print(_stdout.read().decode())
        client.close()

    else:

        # run script on Raspberry Pi to set mask pattern
        command = f"{rpi_python} {script} --file_path {local_path} >/dev/null 2>&1"
        if verbose:
            print(f"COMMAND : {command}")
        os.system(command)

    os.remove(local_path)


def get_programmable_mask(
    vals, sensor, slm_param, rotate=None, flipud=False, nbits=8, color_filter=None, deadspace=True
):
    """
    Get mask as a numpy or torch array. Return same type.

    Parameters
    ----------
    vals : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
        Values to set on programmable mask.
    sensor : :py:class:`~lensless.hardware.sensor.VirtualSensor`
        Sensor object.
    slm_param : dict
        SLM parameters.
    rotate : float, optional
        Rotation angle in degrees.
    flipud : bool, optional
        Flip mask vertically.
    nbits : int, optional
        Number of bits/levels to quantize mask to.
    deadspace: bool, optional
        Whether to include deadspace around mask. Default is True.

    """

    assert waveprop_available

    use_torch = False
    if torch_available:
        use_torch = isinstance(vals, torch.Tensor)
    dtype = vals.dtype

    # -- prepare SLM mask
    n_active_slm_pixels = vals.shape
    n_color_filter = np.prod(slm_param["color_filter"].shape[:2])

    # -- prepare color filter
    if color_filter is None and SLMParam_wp.COLOR_FILTER in slm_param.keys():
        color_filter = slm_param[SLMParam_wp.COLOR_FILTER]
        if isinstance(vals, torch.Tensor):
            color_filter = torch.tensor(color_filter).to(vals)

    if color_filter is not None:

        if isinstance(color_filter, np.ndarray):
            if flipud:
                color_filter = np.flipud(color_filter)
        elif isinstance(color_filter, torch.Tensor):
            if flipud:
                color_filter = torch.flip(color_filter, dims=(0,))
        else:
            raise ValueError("color_filter must be numpy array or torch tensor")

    # -- prepare mask
    if use_torch:
        mask = torch.zeros((n_color_filter,) + tuple(sensor.resolution)).to(vals)
        slm_vals_flat = vals.flatten()
    else:
        mask = np.zeros((n_color_filter,) + tuple(sensor.resolution), dtype=dtype)
        slm_vals_flat = vals.reshape(-1)
    pixel_pitch = slm_param[SLMParam_wp.PITCH]
    d1 = sensor.pitch
    if deadspace:

        centers = get_centers(n_active_slm_pixels, pixel_pitch=pixel_pitch)

        _height_pixel, _width_pixel = (slm_param[SLMParam_wp.CELL_SIZE] / d1).astype(int)

        for i, _center in enumerate(centers):

            _center_pixel = (_center / d1 + sensor.resolution / 2).astype(int)
            _center_top_left_pixel = (
                _center_pixel[0] - np.floor(_height_pixel / 2).astype(int),
                _center_pixel[1] + 1 - np.floor(_width_pixel / 2).astype(int),
            )
            color_filter_idx = i // n_active_slm_pixels[1] % n_color_filter

            mask_val = slm_vals_flat[i] * color_filter[color_filter_idx][0]
            if isinstance(mask_val, np.ndarray):
                mask_val = mask_val[:, np.newaxis, np.newaxis]
            elif isinstance(mask_val, torch.Tensor):
                mask_val = mask_val.unsqueeze(-1).unsqueeze(-1)
            mask[
                :,
                _center_top_left_pixel[0] : _center_top_left_pixel[0] + _height_pixel,
                _center_top_left_pixel[1] : _center_top_left_pixel[1] + _width_pixel,
            ] = mask_val

    else:

        # use color filter to turn mask into RGB
        if use_torch:
            active_mask_rgb = torch.zeros((n_color_filter,) + n_active_slm_pixels).to(vals)
        else:
            active_mask_rgb = np.zeros((n_color_filter,) + n_active_slm_pixels, dtype=dtype)

        # TODO avoid for loop
        for i in range(n_active_slm_pixels[0]):
            row_idx = i % color_filter.shape[0]
            for j in range(n_active_slm_pixels[1]):

                col_idx = j % color_filter.shape[1]
                color_filter_idx = color_filter[row_idx, col_idx]
                active_mask_rgb[
                    :, n_active_slm_pixels[0] - i - 1, n_active_slm_pixels[1] - j - 1
                ] = (vals[i, j] * color_filter_idx)

        # size of active pixels in pixels
        n_active_dim = np.around(slm_param[SLMParam_wp.PITCH] * n_active_slm_pixels / d1).astype(
            int
        )
        # n_active_dim = np.around(slm_param[SLMParam_wp.CELL_SIZE] * n_active_slm_pixels / d1).astype(int)

        # resize to n_active_dim
        if use_torch:
            mask_active = transforms.functional.resize(
                active_mask_rgb, n_active_dim, interpolation=InterpolationMode.NEAREST
            )
        else:
            # TODO check
            mask_active = np.zeros((n_color_filter,) + tuple(n_active_dim), dtype=dtype)
            for i in range(n_color_filter):
                mask_active[i] = np.resize(active_mask_rgb[i], n_active_dim)

        # pad to full mask
        top_left = (sensor.resolution - n_active_dim) // 2
        mask[
            :,
            top_left[0] : top_left[0] + n_active_dim[0],
            top_left[1] : top_left[1] + n_active_dim[1],
        ] = mask_active

    # # quantize mask
    # if use_torch:
    #     mask = mask / torch.max(mask)
    #     mask = torch.round(mask * (2**nbits - 1)) / (2**nbits - 1)
    # else:
    #     mask = mask / np.max(mask)
    #     mask = np.round(mask * (2**nbits - 1)) / (2**nbits - 1)

    # rotate
    if rotate is not None:
        if use_torch:
            mask = transforms.functional.rotate(mask, angle=rotate)
        else:
            mask = rotate_func(mask, axes=(2, 1), angle=rotate, reshape=False)

    return mask


def adafruit_sub2full(
    subpattern,
    center,
):
    sub_shape = subpattern.shape
    controllable_shape = (3, sub_shape[0] // 3, sub_shape[1])
    subpattern_rgb = subpattern.reshape(controllable_shape, order="F")
    subpattern_rgb *= 255

    # pad to full pattern
    pattern = np.zeros((3, 128, 160), dtype=np.uint8)
    top_left = [center[0] - controllable_shape[1] // 2, center[1] - controllable_shape[2] // 2]
    pattern[
        :,
        top_left[0] : top_left[0] + controllable_shape[1],
        top_left[1] : top_left[1] + controllable_shape[2],
    ] = subpattern_rgb.astype(np.uint8)
    return pattern


def full2subpattern(
    pattern,
    shape,
    center,
    slm=None,
):
    shape = np.array(shape)
    center = np.array(center)

    # extract region
    idx_1 = center[0] - shape[0] // 2
    idx_2 = center[1] - shape[1] // 2
    subpattern = pattern[:, idx_1 : idx_1 + shape[0], idx_2 : idx_2 + shape[1]]
    subpattern = subpattern / 255.0
    if slm == "adafruit":
        # flatten color channel along rows
        subpattern = subpattern.reshape((-1, subpattern.shape[-1]), order="F")
    return subpattern


def get_intensity_psf(
    mask,
    waveprop=False,
    sensor=None,
    scene2mask=None,
    mask2sensor=None,
    color_system=None,
):
    """
    Get intensity PSF from mask pattern. Return same type of data.

    Parameters
    ----------
    mask : :py:class:`~numpy.ndarray` or :py:class:`~torch.Tensor`
        Mask pattern.
    waveprop : bool, optional
        Whether to use wave propagation to compute PSF. Default is False,
        namely to return squared intensity of mask pattern as the PSF (i.e.,
        no wave propagation and just shadow of pattern).
    sensor : :py:class:`~lensless.hardware.sensor.VirtualSensor`
        Sensor object. Not used if ``waveprop=False``.
    scene2mask : float
        Distance from scene to mask. Not used if ``waveprop=False``.
    mask2sensor : float
        Distance from mask to sensor. Not used if ``waveprop=False``.
    color_system : :py:class:`~waveprop.color.ColorSystem`, optional
        Color system. Not used if ``waveprop=False``.

    """
    assert waveprop_available

    if color_system is None:
        color_system = ColorSystem.rgb()

    is_torch = False
    device = None
    if torch_available and isinstance(mask, torch.Tensor):
        is_torch = True
        device = mask.device

    dtype = mask.dtype
    ctype, _ = get_ctypes(dtype, is_torch)

    if is_torch:
        psfs = torch.zeros(mask.shape, dtype=ctype, device=device)
    else:
        psfs = np.zeros(mask.shape, dtype=ctype)

    if waveprop:

        assert sensor is not None, "sensor must be specified"
        assert scene2mask is not None, "scene2mask must be specified"
        assert mask2sensor is not None, "mask2sensor must be specified"

        assert (
            len(color_system.wv) == mask.shape[0]
        ), "Number of wavelengths must match number of color channels"

        # spherical wavefronts to mask
        spherical_wavefront = spherical_prop(
            in_shape=sensor.resolution,
            d1=sensor.pitch,
            wv=color_system.wv,
            dz=scene2mask,
            return_psf=True,
            is_torch=is_torch,
            device=device,
            dtype=dtype,
        )
        u_in = spherical_wavefront * mask

        # free space propagation to sensor
        for i, wv in enumerate(color_system.wv):
            psfs[i], _, _ = angular_spectrum(
                u_in=u_in[i],
                wv=wv,
                d1=sensor.pitch,
                dz=mask2sensor,
                dtype=dtype,
                device=device,
            )

    else:

        psfs = mask

    # -- intensity PSF
    if is_torch:
        psf_in = torch.square(torch.abs(psfs))
    else:
        psf_in = np.square(np.abs(psfs))

    return psf_in



# #############################################################################
# trainable_inversion.py
# ======================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################

from lensless.recon.trainable_recon import TrainableReconstructionAlgorithm


class TrainableInversion(TrainableReconstructionAlgorithm):
    def __init__(self, psf, dtype=None, K=1e-4, **kwargs):
        """
        Constructor for trainable inversion component as proposed in
        the FlatNet work: https://siddiquesalman.github.io/flatnet/

        Their implementation: https://github.com/siddiquesalman/flatnet/blob/d1dc179666a08df58c4bdf83c4274310ba3cd186/models/fftlayer.py#L70

        Parameters
        ----------
        psf : :py:class:`~torch.Tensor`
            Point spread function (PSF) that models forward propagation.
            Must be of shape (depth, height, width, channels) even if
            depth = 1 and channels = 1. You can use :py:func:`~lensless.io.load_psf`
            to load a PSF from a file such that it is in the correct format.
        dtype : float32 or float64
            Data type to use for optimization.
        K : float
            Regularization parameter.

        """

        super(TrainableInversion, self).__init__(psf, n_iter=1, dtype=dtype, reset=False, **kwargs)
        self._convolver._Hadj = self._convolver._Hadj / (self._convolver._H.norm() ** 2 + K)

        self.reset()

    def _form_image(self):
        self._image_est[self._image_est < 0] = 0
        return self._image_est

    def _set_psf(self, psf):
        return super()._set_psf(psf)

    def reset(self, batch_size=1):
        # no state variables
        return

    def _update(self, iter):
        self._image_est = self._convolver.deconvolve(self._data)



# #############################################################################
# aperture.py
# =================
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################


from enum import Enum

import numpy as np
from lensless.utils.image import rgb2gray


class ApertureOptions(Enum):
    RECT = "rect"
    SQUARE = "square"
    LINE = "line"
    CIRC = "circ"

    @staticmethod
    def values():
        return [shape.value for shape in ApertureOptions]


class Aperture:
    def __init__(self, shape, pixel_pitch):
        """
        Class for defining VirtualSLM.

        :param shape: (height, width) in number of cell.
        :type shape: tuple(int)
        :param pixel_pitch: Pixel pitch (height, width) in meters.
        :type pixel_pitch: tuple(float)
        """
        assert np.all(shape) > 0
        assert np.all(pixel_pitch) > 0
        self._shape = shape
        self._pixel_pitch = pixel_pitch
        self._values = np.zeros((3,) + shape, dtype=np.uint8)

    @property
    def size(self):
        return np.prod(self._shape)

    @property
    def shape(self):
        return self._shape

    @property
    def pixel_pitch(self):
        return self._pixel_pitch

    @property
    def center(self):
        return np.array([self.height / 2, self.width / 2])

    @property
    def dim(self):
        return np.array(self._shape) * np.array(self._pixel_pitch)

    @property
    def height(self):
        return self.dim[0]

    @property
    def width(self):
        return self.dim[1]

    @property
    def values(self):
        return self._values

    @property
    def grayscale_values(self):
        return rgb2gray(self._values)

    def at(self, physical_coord, value=None):
        """
        Get/set values of VirtualSLM at physical coordinate in meters.

        :param physical_coord: Physical coordinates to get/set VirtualSLM values.
        :type physical_coord: int, float, slice tuples
        :param value: [Optional] values to set, otherwise return values at
            specified coordinates. Defaults to None
        :type value: int, float, :py:class:`~numpy.ndarray`, optional
        :return: If getter is used, values at those coordinates
        :rtype: ndarray
        """
        idx = prepare_index_vals(physical_coord, self._pixel_pitch)
        if value is None:
            # getter
            return self._values[idx]
        else:
            # setter
            self._values[idx] = value

    def __getitem__(self, key):
        return self._values[key]

    def __setitem__(self, key, value):
        self._values[key] = value

    def plot(self, show_tick_labels=False):
        """
        Plot Aperture.

        :param show_tick_labels: Whether to show cell number along x- and y-axis, defaults to False
        :type show_tick_labels: bool, optional
        :return: The axes of the plot.
        :rtype: Axes
        """
        # prepare mask data for `imshow`, expects the input data array size to be (width, height, 3)
        Z = self.values.transpose(1, 2, 0)

        # plot
        import matplotlib.pyplot as plt

        _, ax = plt.subplots()
        extent = [
            -0.5 * self._pixel_pitch[1],
            (self._shape[1] - 0.5) * self._pixel_pitch[1],
            (self._shape[0] - 0.5) * self._pixel_pitch[0],
            -0.5 * self._pixel_pitch[0],
        ]
        ax.imshow(Z, extent=extent)
        ax.grid(which="major", axis="both", linestyle="-", color="0.5", linewidth=0.25)

        x_ticks = np.arange(-0.5, self._shape[1], 1) * self._pixel_pitch[1]
        ax.set_xticks(x_ticks)
        if show_tick_labels:
            x_tick_labels = (np.arange(-0.5, self._shape[1], 1) + 0.5).astype(int)
        else:
            x_tick_labels = [None] * len(x_ticks)
        ax.set_xticklabels(x_tick_labels)

        y_ticks = np.arange(-0.5, self._shape[0], 1) * self._pixel_pitch[0]
        ax.set_yticks(y_ticks)
        if show_tick_labels:
            y_tick_labels = (np.arange(-0.5, self._shape[0], 1) + 0.5).astype(int)
        else:
            y_tick_labels = [None] * len(y_ticks)
        ax.set_yticklabels(y_tick_labels)
        return ax


def rect_aperture(slm_shape, pixel_pitch, apert_dim, center=None):
    """
    Create and return VirtualSLM object with rectangular aperture of desired dimensions.

    :param slm_shape: Dimensions (height, width) of VirtualSLM in cells.
    :type slm_shape: tuple(int)
    :param pixel_pitch: Dimensions (height, width) of each cell in meters.
    :type pixel_pitch: tuple(float)
    :param apert_dim: Dimensions (height, width) of aperture in meters.
    :type apert_dim: tuple(float)
    :param center: [Optional] center of aperture along (SLM) coordinates, indexing starts in top-left corner.
        Default behavior is to place center of aperture at center of SLM.
        Defaults to None
    :type center: tuple(float), optional
    :raises ValueError: If aperture does extend over the boarder of the SLM.
    :return: VirtualSLM object with cells programmed to desired rectangular aperture.
    :rtype: :py:class:`~mask_designer.virtual_slm.VirtualSLM`
    """
    # check input values
    assert np.all(apert_dim) > 0

    # initialize SLM
    slm = Aperture(shape=slm_shape, pixel_pitch=pixel_pitch)

    # check / compute center
    if center is None:
        center = slm.center
    else:
        assert (
            0 <= center[0] < slm.height
        ), f"Center {center} must lie within VirtualSLM dimensions {slm.dim}."
        assert (
            0 <= center[1] < slm.width
        ), f"Center {center} must lie within VirtualSLM dimensions {slm.dim}."

    # compute mask
    apert_dim = np.array(apert_dim)
    top_left = center - apert_dim / 2
    bottom_right = top_left + apert_dim

    if (
        top_left[0] < 0
        or top_left[1] < 0
        or bottom_right[0] >= slm.dim[0]
        or bottom_right[1] >= slm.dim[1]
    ):
        raise ValueError(
            f"Aperture ({top_left[0]}:{bottom_right[0]}, "
            f"{top_left[1]}:{bottom_right[1]}) extends past valid "
            f"VirtualSLM dimensions {slm.dim}"
        )
    slm.at(
        physical_coord=np.s_[top_left[0] : bottom_right[0], top_left[1] : bottom_right[1]],
        value=255,
    )

    return slm


def line_aperture(slm_shape, pixel_pitch, length, vertical=True, center=None):
    """
    Create and return VirtualSLM object with a line aperture of desired length.

    :param slm_shape: Dimensions (height, width) of VirtualSLM in cells.
    :type slm_shape: tuple(int)
    :param pixel_pitch: Dimensions (height, width) of each cell in meters.
    :type pixel_pitch: tuple(float)
    :param length: Length of aperture in meters.
    :type length: float
    :param vertical: Orient line vertically, defaults to True.
    :type vertical: bool, optional
    :param center: [Optional] center of aperture along (SLM) coordinates, indexing starts in top-left corner.
        Default behavior is to place center of aperture at center of SLM.
        Defaults to None
    :type center: tuple(float), optional
    :return: VirtualSLM object with cells programmed to desired line aperture.
    :rtype: :py:class:`~mask_designer.virtual_slm.VirtualSLM`
    """
    # call `create_rect_aperture`
    apert_dim = (length, pixel_pitch[1]) if vertical else (pixel_pitch[0], length)
    return rect_aperture(slm_shape, pixel_pitch, apert_dim, center)


def square_aperture(slm_shape, pixel_pitch, side, center=None):
    """
    Create and return VirtualSLM object with a square aperture of desired shape.

    :param slm_shape: Dimensions (height, width) of VirtualSLM in cells.
    :type slm_shape: tuple(int)
    :param pixel_pitch: Dimensions (height, width) of each cell in meters.
    :type pixel_pitch: tuple(float)
    :param side: Side length of square aperture in meters.
    :type side: float
    :param center: [Optional] center of aperture along (SLM) coordinates, indexing starts in top-left corner.
        Default behavior is to place center of aperture at center of SLM.
        Defaults to None
    :type center: tuple(float), optional
    :return: VirtualSLM object with cells programmed to desired square aperture.
    :rtype: :py:class:`~mask_designer.virtual_slm.VirtualSLM`
    """
    return rect_aperture(slm_shape, pixel_pitch, (side, side), center)


def circ_aperture(slm_shape, pixel_pitch, radius, center=None):
    """
    Create and return VirtualSLM object with a circle aperture of desired shape.

    :param slm_shape: Dimensions (height, width) of VirtualSLM in cells.
    :type slm_shape: tuple(int)
    :param pixel_pitch: Dimensions (height, width) of each cell in meters.
    :type pixel_pitch: tuple(float)
    :param radius: Radius of aperture in meters.
    :type radius: float
    :param center: [Optional] center of aperture along (SLM) coordinates, indexing starts in top-left corner.
        Default behavior is to place center of aperture at center of SLM.
        Defaults to None
    :type center: tuple(float), optional
    :return: VirtualSLM object with cells programmed to desired circle aperture.
    :rtype: :py:class:`~mask_designer.virtual_slm.VirtualSLM`
    """
    # check input values
    assert radius > 0

    # initialize SLM
    slm = Aperture(shape=slm_shape, pixel_pitch=pixel_pitch)

    # check / compute center
    if center is None:
        center = slm.center
    else:
        assert (
            0 <= center[0] < slm.height
        ), f"Center {center} must lie within VirtualSLM dimensions {slm.dim}."
        assert (
            0 <= center[1] < slm.width
        ), f"Center {center} must lie within VirtualSLM dimensions {slm.dim}."

    # compute mask
    i, j = np.meshgrid(
        np.arange(slm.dim[0], step=slm.pixel_pitch[0]),
        np.arange(slm.dim[1], step=slm.pixel_pitch[1]),
        sparse=True,
        indexing="ij",
    )
    x2 = (i - center[0]) ** 2
    y2 = (j - center[1]) ** 2
    slm[:] = 255 * (x2 + y2 < radius**2)
    return slm


def _cell_slice(_slice, cell_m):
    """
    Convert slice indexing in meters to slice indexing in cells.

    author: Eric Bezzam,
    email: ebezzam@gmail.com,
    GitHub: https://github.com/ebezzam

    :param _slice: Original slice in meters.
    :type _slice: slice
    :param cell_m: Dimension of cell in meters.
    :type cell_m: float
    :return: The new slice
    :rtype: slice
    """
    start = None if _slice.start is None else _m_to_cell_idx(_slice.start, cell_m)
    stop = _m_to_cell_idx(_slice.stop, cell_m) if _slice.stop is not None else None
    step = _m_to_cell_idx(_slice.step, cell_m) if _slice.step is not None else None
    return slice(start, stop, step)


def _m_to_cell_idx(val, cell_m):
    """
    Convert location to cell index.

    author: Eric Bezzam,
    email: ebezzam@gmail.com,
    GitHub: https://github.com/ebezzam

    :param val: Location in meters.
    :type val: float
    :param cell_m: Dimension of cell in meters.
    :type cell_m: float
    :return: The cell index.
    :rtype: int
    """
    return int(np.round(val / cell_m))


def prepare_index_vals(key, pixel_pitch):
    """
    Convert indexing object in meters to indexing object in cell indices.

    author: Eric Bezzam,
    email: ebezzam@gmail.com,
    GitHub: https://github.com/ebezzam

    :param key: Indexing operation in meters.
    :type key: int, float, slice, or list
    :param pixel_pitch: Pixel pitch (height, width) in meters.
    :type pixel_pitch: tuple(float)
    :raises ValueError: If the key is of the wrong type.
    :raises NotImplementedError: If key is of size 3, individual channels can't
        be indexed.
    :raises ValueError: If the key has the wrong dimensions.
    :return: The new indexing object.
    :rtype: tuple[slice, int] | tuple[slice, slice] | tuple[slice, ...]
    """
    if isinstance(key, (float, int)):
        idx = slice(None), _m_to_cell_idx(key, pixel_pitch[0])

    elif isinstance(key, slice):
        idx = slice(None), _cell_slice(key, pixel_pitch[0])

    elif len(key) == 2:
        idx = [slice(None)]
        for k, _slice in enumerate(key):

            if isinstance(_slice, slice):
                idx.append(_cell_slice(_slice, pixel_pitch[k]))

            elif isinstance(_slice, (float, int)):
                idx.append(_m_to_cell_idx(_slice, pixel_pitch[k]))

            else:
                raise ValueError("Invalid key.")
        idx = tuple(idx)

    elif len(key) == 3:
        raise NotImplementedError("Cannot index individual channels.")

    else:
        raise ValueError("Invalid key.")
    return idx



import warnings
import hydra
from datetime import datetime
import numpy as np
from slm_controller import slm
from slm_controller.hardware import SLMParam, slm_devices
import matplotlib.pyplot as plt

from lensless.hardware.slm import set_programmable_mask
from lensless.hardware.aperture import rect_aperture, circ_aperture
from lensless.hardware.utils import set_mask_sensor_distance


@hydra.main(version_base=None, config_path="../../configs", config_name="digicam_config")
def config_digicam(config):

    rpi_username = config.rpi.username
    rpi_hostname = config.rpi.hostname
    device = config.device

    shape = slm_devices[device][SLMParam.SLM_SHAPE]
    if not slm_devices[device][SLMParam.MONOCHROME]:
        shape = (3, *shape)
    pixel_pitch = slm_devices[device][SLMParam.PIXEL_PITCH]

    # set mask to sensor distance
    if config.z is not None and not config.virtual:
        set_mask_sensor_distance(config.z, rpi_username, rpi_hostname)

    center = np.array(config.center) * pixel_pitch

    # create random pattern
    pattern = None
    if config.pattern.endswith(".npy"):
        pattern = np.load(config.pattern)
    elif config.pattern == "random":
        rng = np.random.RandomState(1)
        # pattern = rng.randint(low=0, high=np.iinfo(np.uint8).max, size=shape, dtype=np.uint8)
        pattern = rng.uniform(low=config.min_val, high=1, size=shape)
        pattern = (pattern * np.iinfo(np.uint8).max).astype(np.uint8)

    elif config.pattern == "rect":
        rect_shape = config.rect_shape
        apert_dim = rect_shape[0] * pixel_pitch[0], rect_shape[1] * pixel_pitch[1]
        ap = rect_aperture(
            apert_dim=apert_dim,
            slm_shape=slm_devices[device][SLMParam.SLM_SHAPE],
            pixel_pitch=pixel_pitch,
            center=center,
        )
        pattern = ap.values
    elif config.pattern == "circ":
        ap = circ_aperture(
            radius=config.radius * pixel_pitch[0],
            slm_shape=slm_devices[device][SLMParam.SLM_SHAPE],
            pixel_pitch=pixel_pitch,
            center=center,
        )
        pattern = ap.values
    else:
        raise ValueError(f"Pattern {config.pattern} not supported")

    # apply aperture
    if config.aperture is not None:

        # aperture = np.zeros(shape, dtype=np.uint8)
        # top_left = np.array(config.aperture.center) - np.array(config.aperture.shape) // 2
        # bottom_right = top_left + np.array(config.aperture.shape)
        # aperture[:, top_left[0] : bottom_right[0], top_left[1] : bottom_right[1]] = 1

        apert_dim = np.array(config.aperture.shape) * np.array(pixel_pitch)
        ap = rect_aperture(
            apert_dim=apert_dim,
            slm_shape=slm_devices[device][SLMParam.SLM_SHAPE],
            pixel_pitch=pixel_pitch,
            center=np.array(config.aperture.center) * pixel_pitch,
        )
        aperture = ap.values
        aperture[aperture > 0] = 1
        pattern = pattern * aperture

    # save pattern
    if not config.pattern.endswith(".npy") and config.save:
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        pattern_fn = f"{device}_{config.pattern}_pattern_{timestamp}.npy"
        np.save(pattern_fn, pattern)
        print(f"Saved pattern to {pattern_fn}")

    print("Pattern shape : ", pattern.shape)
    print("Pattern dtype : ", pattern.dtype)
    print("Pattern min   : ", pattern.min())
    print("Pattern max   : ", pattern.max())

    assert pattern is not None

    n_nonzero = np.count_nonzero(pattern)
    print(f"Nonzero pixels: {n_nonzero}")

    if not config.virtual:
        set_programmable_mask(pattern, device, rpi_username, rpi_hostname)

    # preview mask
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        s = slm.create(device)
        if config.preview:
            s._show_preview(pattern)
        plt.savefig("preview.png")


if __name__ == "__main__":
    config_digicam()



# #############################################################################
# metric.py
# =========
# Authors :
# Eric BEZZAM [ebezzam@gmail.com]
# #############################################################################


"""
Evaluation
==========

After performing reconstruction, we are typically interested in evaluating the
quality of the reconstruction. To this end, four metrics are made available:

* **Mean squared error (MSE)**: lower is better with a minimum of 0.
* **Peak signal-to-noise ratio (PSNR)**: higher is better with values given in decibels (dB).
* **Structural similarity index measure (SSIM)**: higher is better with a maximum of 1.
* **Learned Perceptual Image Patch Similarity (LPIPS)**: perceptual metrics that used a pre-trained neural network on patches. Lower is better with a minimum of 0. *NB: only for RGB!*

Note that in the examples below, YAML configuration files are read from the ``configs`` directory.
``--help`` can be used to see the available options.

On a single file
----------------

The script ``scripts/compute_metrics_from_original.py`` shows how to compute the above metrics for a
single file by (1) extraction a region of interest and (2) comparing it to a reference file.

After downloading the example files:

.. code:: bash

    wget https://drive.switch.ch/index.php/s/NdgHlcDeHVDH5ww/download -O data.zip
    unzip data.zip -d data
    cp -r data/*/* data/
    rm -rf data/LenslessPiCam_GitHub_data
    rm data.zip

The script can be run with:

.. code:: bash

    python scripts/compute_metrics_from_original.py


Default parameters will be used from the ``configs/compute_metrics_from_original.yaml`` file.

More information can be found in
`this Medium article <https://medium.com/@bezzam/image-similarity-metrics-applied-to-diffusercam-21998967af8d>`__.


DiffuserCam Lensless Mirflickr Dataset (DLMD)
---------------------------------------------

The `DiffuserCam Lensless Mirflickr Dataset (DLMD) <https://waller-lab.github.io/LenslessLearning/dataset.html>`__
comes with (lensed, lensless) image pairs, namely an image that is captured
with a conventional lensed camera and a corresponding image that is captured
with the diffuser-based camera.

The original dataset is quite large (25000 files, 100 GB). So we've prepared
`this subset <https://drive.switch.ch/index.php/s/vmAZzryGI8U8rcE>`__
(200 files, 725 MB).

After downloading the data, you can run ADMM on the subset with the following script.

.. code:: bash

   python scripts/evaluate_mirflickr_admm.py

The default parameters can be found in the ``configs/evaluate_mirflickr_admm.yaml`` file.

It is also possible to set the number of files.

.. code:: bash

   python scripts/evaluate_mirflickr_admm.py n_files=10 save=True


The ``save`` option will save a viewable image for each reconstruction.

You can also apply ADMM on a single image and visualize the iterative
reconstruction.

.. code:: bash

   python scripts/apply_admm_single_mirflickr.py

The default parameters can be found in the ``configs/apply_admm_single_mirflickr.yaml`` file.


Benchmarking with PyTorch
-------------------------

It may be useful to benchmark reconstruction algorithms with PyTorch, e.g.
with a *parallel* dataset of lensless and corresponding lensed images.

:py:class:`~lensless.benchmark.ParallelDataset` is a PyTorch :py:class:`~torch.utils.data.Dataset` object that can be used
to load a parallel dataset of lensless and corresponding lensed images.
The function :py:func:`~lensless.benchmark.benchmark` can be used to evaluate a reconstruction
algorithm on a parallel dataset in batches.

Running the following file will evaluate ADMM on a subset of DLMD:

.. code:: bash

    python lensless/benchmark.py

"""

import numpy as np
from skimage.metrics import mean_squared_error, peak_signal_noise_ratio, structural_similarity
import lpips as lpips_lib
import torch
from scipy.ndimage import rotate
from lensless.utils.image import resize


def mse(true, est, normalize=True):
    """
    Compute the mean-squared error between two images. The closer to 0, the
    closer the match.

    Parameters
    ----------
    true : :py:class:`~numpy.ndarray`
        Ground-truth image, same shape as `est`.
    est : :py:class:`~numpy.ndarray`
        Test image.
    normalize : bool
        Whether to normalize such that maximum value is 1.

    Returns
    -------
    mse : float
        The mean-squared error (MSE) metric.

    """
    if normalize:
        true = np.array(true, dtype=np.float32)
        est = np.array(est, dtype=np.float32)
        true /= true.max()
        est /= est.max()
    return mean_squared_error(image0=true, image1=est)


def psnr(true, est, normalize=True):
    """
    Compute the peak signal to noise ratio (PSNR) for an image. The higher the
    value, the better the match.

    Parameters
    ----------
    true : :py:class:`~numpy.ndarray`
        Ground-truth image, same shape as `est`.
    est : :py:class:`~numpy.ndarray`
        Test image.
    normalize : bool
        Whether to normalize such that maximum value is 1.

    Returns
    -------
    psnr : float
        The PSNR metric.

    """
    if normalize:
        true = np.array(true, dtype=np.float32)
        est = np.array(est, dtype=np.float32)
        true /= true.max()
        est /= est.max()
    return peak_signal_noise_ratio(image_true=true, image_test=est)


def ssim(true, est, normalize=True, channel_axis=2, data_range=None, **kwargs):
    """
    Compute the mean structural similarity index between two images. Values lie
    within [0, 1]. The closer to 1, the closer the match.

    Parameters
    ----------
    true : :py:class:`~numpy.ndarray`
        Ground-truth image, same shape as `est`.
    est : :py:class:`~numpy.ndarray`
        Test image.
    normalize : bool, optional
        Whether to normalize such that maximum value is 1.
    channel_axis : int or None, optional
        If `None`, the image is assumed to be a grayscale (single channel) image.
        Otherwise, this parameter indicates which axis of the array corresponds
        to channels.
    data_range : float or None, optional
        The data range of the input image (distance between minimum and maximum
        possible values). By default, this is estimated from the image data-type.

    Returns
    -------
    ssim : float
        The mean structural similarity index over the image.

    """
    if normalize:
        true = np.array(true, dtype=np.float32)
        est = np.array(est, dtype=np.float32)
        true /= true.max()
        est /= est.max()

    if data_range is None:
        # recommended to explictly pass data range
        data_range = true.max() - true.min()
    return structural_similarity(
        im1=true, im2=est, channel_axis=channel_axis, data_range=data_range, **kwargs
    )


LPIPS_MIN_DIM = 31


def lpips(true, est, normalize=True):
    """
    Compute a perceptual metric (LPIPS) between two images. Values lie within
    [0, 1]. The closer to 0, the closer the match.

    GitHub: https://github.com/richzhang/PerceptualSimilarity

    Parameters
    ----------
    true : :py:class:`~numpy.ndarray`
        Ground-truth image, same shape as `est`.
    est : :py:class:`~numpy.ndarray`
        Test image.
    normalize : bool
        Whether to normalize such that maximum value is 1.

    Returns
    -------
    lpips : float
        The LPIPS metric.

    """
    if np.min(true.shape[:2]) < LPIPS_MIN_DIM:
        raise ValueError(
            f"LPIPS requires images to be at least {LPIPS_MIN_DIM}x{LPIPS_MIN_DIM} pixels."
        )
    if normalize:
        true = np.array(true, dtype=np.float32)
        est = np.array(est, dtype=np.float32)
        true /= true.max()
        est /= est.max()
    loss_fn = lpips_lib.LPIPS(net="alex", verbose=False)
    true = torch.from_numpy(
        np.transpose(true, axes=(2, 0, 1))[
            np.newaxis,
        ].copy()
    )
    est = torch.from_numpy(
        np.transpose(est, axes=(2, 0, 1))[
            np.newaxis,
        ].copy()
    )
    return loss_fn.forward(true, est).squeeze().item()


def extract(
    estimate, original, vertical_crop=None, horizontal_crop=None, rotation=0, verbose=False
):
    """
    Utility function to extract matching region in reconstruction and in original
    image. Later will also be resized to the same dimensions as the estimate.

    Parameters
    ----------
    estimate : :py:class:`~numpy.ndarray`
        Reconstructed image from lensless data.
    original : :py:class:`~numpy.ndarray`
        Original image.
    vertical_crop : (int, int)
        Vertical region (in pixels) to keep.
    horizontal_crop : (int, int)
        Horizontal region (in pixels) to keep.
    rotation : float
        Degrees to rotate reconstruction.
    verbose : bool
        Whether to print extracted and resized shapes.

    Returns
    -------
    estimate : :py:class:`~numpy.ndarray`
        Cropped and rotated image estimate.
    img_resize : :py:class:`~numpy.ndarray`
        Original image resized that the dimensions of `estimate`.

    """

    if vertical_crop is None:
        vertical_crop = (0, estimate.shape[0])
    if horizontal_crop is None:
        horizontal_crop = (0, estimate.shape[1])

    # crop and rotate estimate image
    if rotation:
        estimate = rotate(
            estimate[vertical_crop[0] : vertical_crop[1], horizontal_crop[0] : horizontal_crop[1]],
            angle=rotation,
            mode="nearest",
            reshape=False,
        )
    estimate /= estimate.max()
    estimate = np.clip(estimate, 0, 1)
    if verbose:
        print("estimate cropped: ")
        print(estimate.shape)
        print(estimate.dtype)
        print(estimate.max())

    # resize original image accordingly
    factor = estimate.shape[1] / original.shape[1]
    if verbose:
        print("resize factor", factor)
    img_resize = np.zeros_like(estimate)
    tmp = resize(original, factor=factor).astype(estimate.dtype)
    img_resize[
        : min(estimate.shape[0], tmp.shape[0]), : min(estimate.shape[1], tmp.shape[1])
    ] = tmp[: min(estimate.shape[0], tmp.shape[0]), : min(estimate.shape[1], tmp.shape[1])]
    if verbose:
        print("\noriginal resized: ")
        print(img_resize.shape)
        print(img_resize.dtype)
        print(img_resize.max())

    return estimate, img_resize
